"""Defines a web-view tab."""


from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import HSplit, VSplit
from prompt_toolkit.layout.dimension import Dimension
from upath import UPath

from euporie.core.convert.core import get_mime
from euporie.core.current import get_app
from euporie.core.data_structures import DiBool
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.tabs.base import Tab
from euporie.core.utils import run_in_thread_with_context
from euporie.core.widgets.decor import FocusedStyle
from euporie.core.widgets.display import DisplayWindow
from euporie.core.widgets.forms import Button, Text
from euporie.web.widgets.webview import WebViewControl

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app import BaseApp
    from euporie.core.widgets.status_bar import StatusBarFields

log = logging.getLogger(__name__)


class WebTab(Tab):
    """Tab class for displaying files."""

    name = "Web Viewer"
    weight = 2
    mime_types = {"text/html", "text/markdown"}

    def __init__(self, app: BaseApp, path: Path | None) -> None:
        """Call when the tab is created."""
        super().__init__(app, path)
        self.status: Callable[[], StatusBarFields] | None = None

        def _load() -> None:
            self.container = self.load_container()

        run_in_thread_with_context(_load)

    @property
    def title(self) -> str:
        """Return the tab title."""
        if webview := getattr(self, "webview", None):
            if title := webview.title:
                return title
        if self.path is not None:
            return str(self.path.name) or str(self.path)
        else:
            return "<Web Page>"

    def load_url(self, url: str | Path, new_tab: bool = False, **kwargs: Any) -> bool:
        """Load a new URL, or the URL in the address-bar."""
        log.debug("Loading %s", url)
        if not url:
            return False
        if isinstance(url, str):
            url = UPath(url)
        if not new_tab and get_mime(url) in self.mime_types:
            self.webview.load_url(url, **kwargs)
        else:
            get_app().open_file(url)
        return True

    def _url_loaded(self, webview: WebViewControl) -> None:
        """Trigger callback when the URL is loaded."""
        url = webview.url
        if url is not None:
            self.path = UPath(url)
            self.url_bar.text = str(url)

    def load_container(self) -> AnyContainer:
        """Abcract method for loading the notebook's main container."""
        assert self.path is not None
        path = self.path
        self.webview = WebViewControl(url=path, link_handler=self.load_url)
        self.webview.rendered += self._url_loaded

        def _status() -> StatusBarFields:
            return (self.webview.status if self.webview is not None else [], [])

        self.status = _status

        button_prev = Button(
            "◀",
            show_borders=DiBool(top=True, right=False, bottom=True, left=True),
            disabled=Condition(lambda: not self.webview.prev_stack),
            on_click=lambda x: self.webview.nav_prev(),
        )
        button_next = Button(
            "▶",
            show_borders=DiBool(top=True, right=True, bottom=True, left=False),
            disabled=Condition(lambda: not self.webview.next_stack),
            on_click=lambda x: self.webview.nav_next(),
        )
        self.url_bar = Text(
            text=str(path),
            show_borders=DiBool(top=True, right=False, bottom=True, left=True),
            accept_handler=lambda buf: self.load_url(buf.text),
        )
        button_go = Button(
            "➜",
            show_borders=DiBool(top=True, right=True, bottom=True, left=False),
            on_click=lambda x: (self.load_url(self.url_bar.text) and None)
            or None,  # Magical typing magic
        )

        return HSplit(
            [
                VSplit(
                    [
                        FocusedStyle(button_prev),
                        FocusedStyle(button_next),
                        FocusedStyle(self.url_bar),
                        FocusedStyle(button_go),
                    ],
                    style="class:tab.toolbar",
                ),
                VSplit(
                    [
                        window := DisplayWindow(self.webview),
                        MarginContainer(ScrollbarMargin(), target=window),
                    ]
                ),
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
        )

    def __pt_status__(self) -> StatusBarFields:
        """Return a list of statusbar field values shown then this tab is active."""
        if callable(self.status):
            return self.status()
        else:
            return ([], [])



"""Relating to lexers."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pygments.lexers import (
    get_lexer_by_name,
    get_lexer_for_filename,
    guess_lexer,
    guess_lexer_for_filename,
)
from pygments.util import ClassNotFound

if TYPE_CHECKING:
    from pathlib import Path

    from pygments.lexer import Lexer as PygmentsLexerCls


def detect_lexer(
    text: str = "", path: Path | None = None, language: str = ""
) -> PygmentsLexerCls | None:
    """Detect the pygments lexer for a file."""
    lexer = None
    if path is not None:
        try:
            lexer = get_lexer_for_filename(path)
        except ClassNotFound:
            try:
                lexer = guess_lexer_for_filename(path, text)
            except ClassNotFound:
                pass
    if lexer is None and language:
        try:
            lexer = get_lexer_by_name(language)
        except ClassNotFound:
            pass
    if lexer is None:
        try:
            lexer = guess_lexer(text)
        except ClassNotFound:
            pass
    return lexer



"""A notebook which renders cells one cell at a time."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    VSplit,
    Window,
)
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.widgets import Box

from euporie.core.config import add_setting
from euporie.core.tabs.notebook import BaseNotebook
from euporie.core.widgets.cell import Cell
from euporie.core.widgets.page import PrintingContainer

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.application.application import Application
    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app import BaseApp

log = logging.getLogger(__name__)


class PreviewNotebook(BaseNotebook):
    """A notebook tab which renders cells sequentially."""

    bg_init = False

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        use_kernel_history: bool = False,
    ) -> None:
        """Create a new instance."""
        self.cell_index = 0
        self.cells: FastDictCache[tuple[int], Cell] = FastDictCache(
            get_value=self.get_cell
        )

        super().__init__(app, path, use_kernel_history=use_kernel_history)

        self.app.before_render += self.before_render
        self.app.after_render += self.after_render

    def pre_init_kernel(self) -> None:
        """Filter cells before kernel is loaded."""
        super().pre_init_kernel()

        # Filter the cells to be shown
        n_cells = len(self.json["cells"]) - 1
        start: int | None = None
        stop: int | None = None
        if self.app.config.cell_start is not None:
            start = min(max(self.app.config.cell_start, -n_cells), n_cells)
        if self.app.config.cell_stop is not None:
            stop = min(max(self.app.config.cell_stop, -n_cells), n_cells)
        log.debug("Showing cells %s to %s", start, stop)
        self.json["cells"] = self.json["cells"][start:stop]

    def post_init_kernel(self) -> None:
        """Optionally start kernel after it is loaded."""
        super().post_init_kernel()
        # If we are running the notebook, pause rendering util the kernel has started
        if self.app.config.run:
            self.app.pause_rendering()
            self.kernel.start(cb=self.kernel_started, wait=True)

    def print_title(self) -> None:
        """Print a notebook's filename."""
        from euporie.core.border import DoubleLine
        from euporie.core.ft.utils import (
            FormattedTextAlign,
            add_border,
            align,
            wrap,
        )

        width = self.app.output.get_size().columns
        ft: StyleAndTextTuples = [("bold", str(self.path))]
        ft = wrap(ft, width - 4)
        ft = align(ft, how=FormattedTextAlign.CENTER, width=width - 4)
        ft = add_border(ft, width=width, border_grid=DoubleLine.grid)
        ft.append(("", "\n"))
        self.app.print_text(ft)

    def kernel_started(self, result: dict | None = None) -> None:
        """Resume rendering the app when the kernel has started."""
        self.app.resume_rendering()

    def close(self, cb: Callable | None = None) -> None:
        """Clean up render hooks before the tab is closed."""
        self.app.after_render -= self.after_render
        if self.app.config.run and self.app.config.save:
            self.save()
        super().close(cb)

    def before_render(self, app: Application[Any]) -> None:
        """Run the cell before rendering it if needed."""
        if (
            self.app.tab == self
            and self.cell_index == 0
            and self.app.config.show_filenames
        ):
            self.print_title()

        if not self.json["cells"]:
            log.error("No cells")
            self.app.print_text([("", "(No cells to display)\n")])
            self.app.close_tab(self)

        elif self.app.config.run:
            cell = self.cell
            cell.run_or_render(wait=True)
            # self.kernel.wait_for_status("idle")

    def after_render(self, app: Application[Any]) -> None:
        """Close the tab if all cells have been rendered."""
        if self.app.tab == self:
            if self.cell_index < len(self.json["cells"]) - 1:
                self.cell_index += 1
            else:
                self.app.close_tab(self)

    def get_cell(self, index: int) -> Cell:
        """Render a cell by its index."""
        if index < len(self.json["cells"]):
            return Cell(index, self.json["cells"][index], self)
        else:
            return Cell(0, {}, self)

    @property
    def cell(self) -> Cell:
        """Return the current cell."""
        return self.cells[(self.cell_index,)]

    def load_container(self) -> AnyContainer:
        """Abcract method for loading the notebook's main container."""
        return PrintingContainer(
            [
                VSplit(
                    [
                        ConditionalContainer(
                            Window(), filter=~self.app.config.filter("expand")
                        ),
                        Box(
                            body=DynamicContainer(lambda: self.cell),
                            padding=0,
                            width=Dimension(
                                preferred=self.app.config.max_notebook_width
                            ),
                        ),
                        ConditionalContainer(
                            Window(), filter=~self.app.config.filter("expand")
                        ),
                    ]
                )
            ]
        )

    # ################################### Settings ####################################

    add_setting(
        name="run",
        flags=["--run"],
        type_=bool,
        help_="Run the notebook files when loaded",
        default=False,
        description="""
            If set, notebooks will be run automatically when opened, or if previewing a
            file, the notebooks will be run before being output.
        """,
    )

    add_setting(
        name="save",
        flags=["--save"],
        type_=bool,
        help_="Save the notebook after running it",
        default=False,
        description="""
            If set, notebooks will be saved after they have been run. This setting only
            has any affect if the :option:`run` setting is active.
        """,
    )

    add_setting(
        name="show_filenames",
        flags=["--show-filenames"],
        type_=bool,
        help_="Show the notebook filenames when previewing multiple notebooks",
        default=False,
        description="""
            If set, the notebook filenames will be printed above each notebook's output
            when multiple notebooks are being previewed.
        """,
    )

    add_setting(
        name="cell_start",
        flags=["--cell-start"],
        type_=int,
        help_="The first cell to include in the preview",
        default=None,
        description="""
            When set, only cells after the given cell index will be shown.
        """,
    )

    add_setting(
        name="cell_stop",
        flags=["--cell-stop"],
        type_=int,
        help_="The last cell to include in the preview",
        default=None,
        description="""
            When set, only cells before the given cell index will be shown.
        """,
    )



"""Miscellaneou utility classes."""

from __future__ import annotations

import contextvars
from itertools import chain
from threading import Thread
from typing import TYPE_CHECKING, Sequence, TypeVar, overload

from prompt_toolkit.mouse_events import MouseButton, MouseEventType

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.mouse_handlers import MouseHandler
    from prompt_toolkit.mouse_events import MouseEvent

T = TypeVar("T")


class ChainedList(Sequence[T]):
    """A list-like class which chains multiple lists."""

    def __init__(self, *lists: Iterable[T]) -> None:
        """Create a new instance."""
        self.lists = lists

    @property
    def data(self) -> list[T]:
        """Return the list data."""
        return list(chain.from_iterable(self.lists))

    @overload
    def __getitem__(self, i: int) -> T:
        ...

    @overload
    def __getitem__(self, i: slice) -> list[T]:
        ...

    def __getitem__(self, i):
        """Get an item from the chained lists."""
        return self.data[i]

    def __len__(self) -> int:
        """Return the length of the chained lists."""
        return len(self.data)


def dict_merge(target_dict: dict, input_dict: dict) -> None:
    """Merge the second dictionary onto the first."""
    for k in input_dict:
        if k in target_dict:
            if isinstance(target_dict[k], dict) and isinstance(input_dict[k], dict):
                dict_merge(target_dict[k], input_dict[k])
            elif isinstance(target_dict[k], list) and isinstance(input_dict[k], list):
                target_dict[k] = [*target_dict[k], *input_dict[k]]
            else:
                target_dict[k] = input_dict[k]
        else:
            target_dict[k] = input_dict[k]


def on_click(func: Callable) -> MouseHandler:
    """Return a mouse handler which call a given function on click."""

    def _mouse_handler(mouse_event: MouseEvent) -> NotImplementedOrNone:
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_UP
        ):
            return func()
        return NotImplemented

    return _mouse_handler


def run_in_thread_with_context(
    func: Callable, *args: Any, daemon: bool = True, **kwargs: Any
) -> None:
    """Run a function in an thread, but make sure it uses the same contextvars.

    This is required so that the function will see the right application.
    """
    Thread(
        target=contextvars.copy_context().run,
        args=(func, *args),
        kwargs=kwargs,
        daemon=daemon,
    ).start()



"""Define input history loaders."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.history import History

if TYPE_CHECKING:
    from typing import AsyncGenerator, Iterable

    from euporie.core.kernel import Kernel

log = logging.getLogger(__name__)


class KernelHistory(History):
    """Load the kernel's command history."""

    def __init__(self, kernel: Kernel, n: int = 1000) -> None:
        """Create a new instance of the kernel history loader."""
        super().__init__()
        self.kernel = kernel
        # How many items to load
        self.n = n
        self.n_loaded = 0
        self.loading = False

    async def load(self) -> AsyncGenerator[str, None]:
        """Load the history and yield all entries, most recent history first.

        This method can be called multiple times from the `Buffer` to
        repopulate the history when prompting for a new input. So we are
        responsible here for both caching, and making sure that strings that
        were were appended to the history will be incorporated next time this
        method is called.

        Yields:
            Each history string
        """
        if not self.loading and not self._loaded and self.kernel.kc:
            self.loading = True
            items = await self.kernel.history_(n=self.n, hist_access_type="tail")
            if items:
                self._loaded_strings = [item[2] for item in reversed(items)]
                # Remove sequential duplicates
                self._loaded_strings = [
                    v
                    for i, v in enumerate(self._loaded_strings)
                    if i == 0 or v != self._loaded_strings[i - 1]
                ]
                self._loaded = True
                self.n_loaded = len(self._loaded_strings)
                log.debug(
                    "Loaded %s items from kernel history", len(self._loaded_strings)
                )
            self.loading = False

        for item in self._loaded_strings:
            yield item

    def load_history_strings(self) -> Iterable[str]:
        """Not used to load history, as we load it asynchronously."""
        while False:
            yield

    def store_string(self, string: str) -> None:
        """Don't store strings in persistent storage: they are stored by the kernel."""
        pass

    @property
    def recent(self) -> list[str]:
        """Return new items added since history was initially loaded."""
        return self._loaded_strings[: -self.n_loaded]



"""Define a cell object with input are and rich outputs, and related objects."""

from __future__ import annotations

import logging
import os
import weakref
from functools import partial
from typing import TYPE_CHECKING, cast

import nbformat
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.lexers import DynamicLexer, PygmentsLexer, SimpleLexer
from pygments.lexers import get_lexer_by_name
from pygments.util import ClassNotFound

from euporie.core.border import NoLine, ThickLine, ThinLine
from euporie.core.config import add_setting
from euporie.core.current import get_app
from euporie.core.filters import multiple_cells_selected
from euporie.core.format import format_code
from euporie.core.utils import on_click
from euporie.core.widgets.cell_outputs import CellOutputArea
from euporie.core.widgets.inputs import KernelInput, StdInput

if TYPE_CHECKING:
    from typing import Any, Callable, Literal

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple
    from prompt_toolkit.lexers import Lexer

    from euporie.core.tabs.notebook import BaseNotebook


log = logging.getLogger(__name__)


def get_cell_id(cell_json: dict) -> str:
    """Return the cell ID field defined in a cell JSON object.

    If no cell ID is defined (as per ```:mod:`nbformat`<4.5``), then one is generated
    and added to the cell.

    Args:
        cell_json: The cell's JSON object as a python dictionary

    Returns:
        The ID string

    """
    cell_id = cell_json.get("id", "")
    # Assign a cell id if missing
    if not cell_id:
        cell_json["id"] = cell_id = nbformat.v4.new_code_cell().get("id")
    return cell_id


def get_pygments_lexer(language: str) -> Lexer:
    """Get a pygments lexer class by name, ignoring errors."""
    try:
        lexer_class = get_lexer_by_name(language).__class__
    except ClassNotFound:
        return SimpleLexer()
    else:
        return PygmentsLexer(
            lexer_class,
            sync_from_start=False,
        )


class Cell:
    """A kernel_tab cell element.

    Contains a transparent clickable overlay, which is not displayed when the cell is
    focused.
    """

    input_box: KernelInput

    def __init__(self, index: int, json: dict, kernel_tab: BaseNotebook) -> None:
        """Initiate the cell element.

        Args:
            index: The position of this cell in the kernel_tab
            json: A reference to the cell's json object
            kernel_tab: The kernel_tab instance this cell belongs to

        """
        weak_self = weakref.proxy(self)
        self.index = index
        self.json = json
        self.kernel_tab: BaseNotebook = kernel_tab
        self.rendered = True
        self.clear_outputs_on_output = False

        self.state = "idle"

        show_input = Condition(
            lambda: bool(
                (weak_self.json.get("cell_type") != "markdown")
                | (
                    (weak_self.json.get("cell_type") == "markdown")
                    & ~weak_self.rendered
                )
            )
        )
        show_output = Condition(
            lambda: (
                (weak_self.json.get("cell_type") != "markdown")
                & bool(weak_self.output_json)
                | ((weak_self.json.get("cell_type") == "markdown") & weak_self.rendered)
            )
        )
        # scroll_input = Condition(
        # lambda: bool(
        # (weak_self.json.get("cell_type") == "markdown") & ~weak_self.rendered
        # )
        # )
        show_prompt = Condition(lambda: weak_self.cell_type == "code")
        self.is_code = Condition(lambda: weak_self.json.get("cell_type") == "code")

        self.output_area = CellOutputArea(json=self.output_json, parent=weak_self)

        def on_text_changed(buf: Buffer) -> None:
            """Update cell json when the input buffer has been edited."""
            weak_self._set_input(buf.text)
            weak_self.kernel_tab.dirty = True

        def on_cursor_position_changed(buf: Buffer) -> None:
            """Respond to cursor movements."""
            # Update contextual help
            if weak_self.kernel_tab.app.config.autoinspect and weak_self.is_code():
                weak_self.input_box.inspect()
            else:
                if pager := get_app().pager:
                    pager.hide()

            # Tell the scrolling container to scroll the cursor into view on the next render
            weak_self.kernel_tab.page.scroll_to_cursor = True

        # Now we generate the main container used to represent a kernel_tab cell

        source_hidden = Condition(
            lambda: self.json["metadata"].get("jupyter", {}).get("source_hidden", False)
        )

        self.input_box = KernelInput(
            kernel_tab=self.kernel_tab,
            text=self.input,
            complete_while_typing=self.is_code,
            autosuggest_while_typing=self.is_code,
            wrap_lines=Condition(lambda: weak_self.json.get("cell_type") == "markdown"),
            on_text_changed=on_text_changed,
            on_cursor_position_changed=on_cursor_position_changed,
            lexer=DynamicLexer(
                partial(
                    lambda cell: (
                        get_pygments_lexer(cell.language)
                        if cell.cell_type != "raw"
                        else SimpleLexer()
                    ),
                    weakref.proxy(self),
                )
            ),
            accept_handler=lambda buffer: self.run_or_render() or True,
            focusable=show_input & ~source_hidden,
            tempfile_suffix=self.tempfile_suffix,
        )
        self.input_box.buffer.name = self.cell_type

        def border_char(name: str) -> Callable[..., str]:
            """Return a function which returns the cell border character to display."""

            def _inner() -> str:
                grid = NoLine.grid
                if get_app().config.show_cell_borders or weak_self.selected:
                    if weak_self.focused and multiple_cells_selected():
                        grid = ThickLine.outer
                    else:
                        grid = ThinLine.outer
                return getattr(grid, name.upper())

            return _inner

        self.control = Window(
            FormattedTextControl(
                border_char("TOP_LEFT"),
                focusable=True,
                show_cursor=False,
            ),
            width=1,
            height=0,
            style="class:border",
            always_hide_cursor=True,
        )

        fill = partial(Window, style="class:border")

        # Create textbox for standard input
        def _send_input(buf: Buffer) -> bool:
            return False

        self.stdin_box = StdInput(self.kernel_tab)

        top_border = VSplit(
            [
                self.control,
                ConditionalContainer(
                    content=fill(
                        char=border_char("TOP_MID"),
                        width=lambda: len(weak_self.prompt),
                        height=1,
                    ),
                    filter=show_prompt,
                ),
                ConditionalContainer(
                    content=fill(width=1, height=1, char=border_char("TOP_SPLIT")),
                    filter=show_prompt,
                ),
                fill(char=border_char("TOP_MID"), height=1),
                fill(width=1, height=1, char=border_char("TOP_RIGHT")),
            ],
            height=1,
        )

        input_row = ConditionalContainer(
            VSplit(
                [
                    fill(width=1, char=border_char("MID_LEFT")),
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(
                                lambda: [
                                    (
                                        "",
                                        weak_self.prompt,
                                        on_click(self.toggle_input),
                                    ),
                                    ("", "\n ", lambda e: NotImplemented),
                                ]
                            ),
                            width=lambda: len(weak_self.prompt),
                            height=Dimension(preferred=1),
                            style="class:input,prompt",
                        ),
                        filter=show_prompt,
                    ),
                    ConditionalContainer(
                        fill(width=1, char=border_char("MID_SPLIT")),
                        filter=show_prompt,
                    ),
                    ConditionalContainer(self.input_box, filter=~source_hidden),
                    ConditionalContainer(
                        Window(
                            FormattedTextControl(
                                [
                                    cast(
                                        "OneStyleAndTextTuple",
                                        (
                                            "class:cell,show,inputs",
                                            " … ",
                                            on_click(self.show_input),
                                        ),
                                    )
                                ]
                            )
                        ),
                        filter=source_hidden,
                    ),
                    fill(width=1, char=border_char("MID_RIGHT")),
                ],
            ),
            filter=show_input,
        )
        middle_line = ConditionalContainer(
            content=VSplit(
                [
                    fill(width=1, height=1, char=border_char("SPLIT_LEFT")),
                    ConditionalContainer(
                        content=fill(
                            char=border_char("SPLIT_MID"),
                            width=lambda: len(weak_self.prompt),
                        ),
                        filter=show_prompt,
                    ),
                    ConditionalContainer(
                        content=fill(
                            width=1, height=1, char=border_char("SPLIT_SPLIT")
                        ),
                        filter=show_prompt,
                    ),
                    fill(char=border_char("SPLIT_MID")),
                    fill(width=1, height=1, char=border_char("SPLIT_RIGHT")),
                ],
                height=1,
            ),
            filter=(show_input & show_output) | self.stdin_box.visible,
        )

        outputs_hidden = Condition(
            lambda: self.json["metadata"]
            .get("jupyter", {})
            .get("outputs_hidden", False)
            or self.json["metadata"].get("collapsed", False)
        )

        output_row = ConditionalContainer(
            VSplit(
                [
                    fill(width=1, char=border_char("MID_LEFT")),
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(
                                lambda: [
                                    (
                                        "",
                                        weak_self.prompt,
                                        on_click(self.toggle_output),
                                    ),
                                    ("", "\n ", lambda e: NotImplemented),
                                ],
                            ),
                            width=lambda: len(weak_self.prompt),
                            height=Dimension(preferred=1),
                            style="class:output,prompt",
                        ),
                        filter=show_prompt,
                    ),
                    ConditionalContainer(
                        content=fill(width=1, char=border_char("MID_SPLIT")),
                        filter=show_prompt,
                    ),
                    ConditionalContainer(
                        fill(width=1, char=border_char("MID_MID")),
                        filter=~show_prompt,
                    ),
                    HSplit(
                        [
                            ConditionalContainer(
                                self.output_area,
                                filter=~outputs_hidden,
                            ),
                            ConditionalContainer(
                                Window(
                                    FormattedTextControl(
                                        [
                                            cast(
                                                "OneStyleAndTextTuple",
                                                (
                                                    "class:cell,show,outputs",
                                                    " … ",
                                                    on_click(self.show_output),
                                                ),
                                            ),
                                        ]
                                    )
                                ),
                                filter=outputs_hidden,
                            ),
                            self.stdin_box,
                        ]
                    ),
                    ConditionalContainer(
                        fill(width=1, char=border_char("MID_MID")),
                        filter=~show_prompt,
                    ),
                    fill(width=1, char=border_char("MID_RIGHT")),
                ],
            ),
            filter=show_output | self.stdin_box.visible,
        )
        bottom_border = VSplit(
            [
                fill(width=1, height=1, char=border_char("BOTTOM_LEFT")),
                ConditionalContainer(
                    content=fill(
                        char=border_char("BOTTOM_MID"),
                        width=lambda: len(weak_self.prompt),
                    ),
                    filter=show_prompt,
                ),
                ConditionalContainer(
                    content=fill(width=1, height=1, char=border_char("BOTTOM_SPLIT")),
                    filter=show_prompt,
                ),
                fill(char=border_char("BOTTOM_MID")),
                fill(width=1, height=1, char=border_char("BOTTOM_RIGHT")),
            ],
            height=1,
        )

        def _style() -> str:
            """Calculate the cell's style given its state."""
            style = "class:cell"
            if weak_self.selected:
                if weak_self.kernel_tab.in_edit_mode():
                    style += ",edit"
                else:
                    style += ",cell.selection"
            return style

        self.container = HSplit(
            [
                top_border,
                input_row,
                middle_line,
                output_row,
                bottom_border,
            ],
            style=_style,
        )

    def focus(self, position: int | None = None, scroll: bool = False) -> None:
        """Focus the relevant control in this cell.

        Args:
            position: An optional cursor position index to apply to the input box
            scroll: Whether to scroll the page to make the selection visible

        """
        to_focus = None
        if self.kernel_tab.edit_mode:
            # Select just this cell when editing
            # self.kernel_tab.select(self.index)
            if self.stdin_box.visible():
                to_focus = self.stdin_box.window
            else:
                self.show_input()
                to_focus = self.input_box.window
                self.rendered = False
            if position is not None:
                self.input_box.buffer.cursor_position = position % (
                    len(self.input_box.buffer.text) or 1
                )
        else:
            to_focus = self.kernel_tab.cell.control

        # We force focus here, bypassing the layout's checks, as the control we want to
        # focus might be not be in the current layout yet.
        get_app().layout._stack.append(to_focus)

        # Scroll the currently selected slice into view
        if scroll:
            self.kernel_tab.scroll_to(self.index)

    @property
    def focused(self) -> bool:
        """Determine if the cell currently has focus."""
        if self.container is not None:
            return get_app().layout.has_focus(self)
        else:
            return False

    @property
    def selected(self) -> bool:
        """Determine if the cell currently is selected."""
        if self.container is not None:
            return self.index in self.kernel_tab.selected_indices
        else:
            return False

    @property
    def cell_type(self) -> str:
        """Determine the current cell type."""
        return self.json.get("cell_type", "code")

    @property
    def tempfile_suffix(self) -> str:
        """Return the file suffix matching the current cell type."""
        cell_type = self.cell_type
        if cell_type == "markdown":
            return ".md"
        elif cell_type == "raw":
            return ""
        else:
            return self.kernel_tab.kernel_lang_file_ext

    @property
    def execution_count(self) -> str:
        """Retrieve the execution count from the cell's JSON."""
        return self.json.get("execution_count", " ")

    @execution_count.setter
    def execution_count(self, count: int) -> None:
        """Set the execution count in the cell's JSON.

        Args:
            count: The new execution count number.

        """
        self.json["execution_count"] = count

    @property
    def prompt(self) -> str:
        """Determine what should be displayed in the prompt of the cell."""
        if self.state in ("busy", "queued"):
            prompt = "*"
        else:
            prompt = self.execution_count or " "
        if prompt:
            prompt = f"[{prompt}]"
        return prompt

    def _set_input(self, value: str) -> None:
        self.json["source"] = value

    @property
    def input(self) -> str:
        """Fetch the cell's contents from the cell's JSON."""
        return self.json.get("source", "")

    @input.setter
    def input(self, value: str) -> None:
        """Set the cell's contents in the cell's JSON.

        Args:
            value: The new cell contents text.

        """
        cp = self.input_box.buffer.cursor_position
        self._set_input(value)
        self.input_box.text = self.json["source"]
        cp = max(0, min(cp, len(value)))
        self.input_box.buffer.cursor_position = cp

    @property
    def output_json(self) -> list[dict[str, Any]]:
        """Retrieve a list of cell outputs from the cell's JSON."""
        if self.cell_type == "markdown":
            return [
                {"data": {"text/x-markdown": self.input}, "output_type": "markdown"}
            ]
        else:
            return self.json.setdefault("outputs", [])

    def refresh(self, now: bool = True) -> None:
        """Request that the cell to be re-rendered next time it is drawn."""
        self.kernel_tab.refresh_cell(self)
        if now:
            get_app().invalidate()

    def ran(self, content: dict | None = None) -> None:
        """Update the cell status and update display when the cell has finished."""
        self.state = "idle"
        self.refresh()

    def remove_outputs(self) -> None:
        """Remove all outputs from the cell."""
        self.clear_outputs_on_output = False
        if "outputs" in self.json:
            self.json["outputs"].clear()
        if self.cell_type != "markdown":
            self.output_area.reset()
        # Ensure the cell output area is visible
        self.show_output()
        self.refresh()

    def set_cell_type(
        self, cell_type: Literal["markdown", "code", "raw"], clear: bool = False
    ) -> None:
        """Convert the cell to a different cell type.

        Args:
            cell_type: The desired cell type.
            clear: If True, cell outputs will be cleared

        """
        if clear and cell_type != "markdown":
            self.remove_outputs()
        if cell_type == "code":
            self.json.setdefault("execution_count", None)
        if cell_type == "markdown" and "execution_count" in self.json:
            del self.json["execution_count"]
        # Record the new cell type
        self.json["cell_type"] = cell_type
        self.input_box.buffer.name = cell_type
        # Update the output-area
        # self.output_area.json = self.output_json
        # Force the input box lexer to re-run
        self.input_box.control._fragment_cache.clear()

    @property
    def id(self) -> str:
        """Return the cell's ID as per the cell JSON."""
        return get_cell_id(self.json)

    @property
    def language(self) -> str:
        """Return the cell's code language."""
        if self.cell_type == "markdown":
            return "markdown"
        elif self.cell_type == "code":
            lang_info = self.kernel_tab.metadata.get("language_info", {})
            return lang_info.get("name", lang_info.get("pygments_lexer", "python"))
        else:
            return "raw"

    def reformat(self) -> None:
        """Reformat the cell's input."""
        config = get_app().config
        self.input = format_code(self.input, config)
        self.refresh()

    def run_or_render(
        self,
        buffer: Buffer | None = None,
        wait: bool = False,
        callback: Callable[..., None] | None = None,
    ) -> bool:
        """Send the cell's source code the the kernel to run.

        Args:
            buffer: Unused parameter, required when accepting the contents of a cell's
                input buffer
            wait: Has no effect
            callback: Callable to run when the kernel has finished running the cell

        Returns:
            Always returns True

        """
        if self.cell_type == "markdown":
            self.output_area.json = self.output_json
            self.rendered = True

        elif self.cell_type == "code":
            if get_app().config.autoformat:
                self.reformat()
            self.state = "queued"
            self.refresh()
            self.kernel_tab.run_cell(self, wait=wait, callback=callback)

        return True

    def __pt_container__(self) -> Container:
        """Return the container which represents this cell."""
        return self.container

    def set_execution_count(self, n: int) -> None:
        """Set the execution count of the cell."""
        self.json["execution_count"] = n

    def add_output(self, output_json: dict[str, Any]) -> None:
        """Add a new output to the cell."""
        # Clear the output if we were previously asked to
        if self.clear_outputs_on_output:
            self.remove_outputs()
        self.json.setdefault("outputs", []).append(output_json)
        # Add the new output to the output area
        self.output_area.add_output(output_json)
        # Tell the page this cell has been updated
        self.refresh()

    def clear_output(self, wait: bool = False) -> None:
        """Remove the cells output, optionally when new output is generated."""
        if wait:
            self.clear_outputs_on_output = True
        else:
            self.remove_outputs()

    def show_input(self) -> None:
        """Set the cell inputs to visible."""
        self.set_metadata(("jupyter", "source_hidden"), False)

    def hide_input(self) -> None:
        """Set the cell inputs to visible."""
        # Exit edit mode
        self.kernel_tab.edit_mode = False
        # Un-focus the cell input
        self.focus()
        # Set the input to hidden
        self.set_metadata(("jupyter", "source_hidden"), True)

    def toggle_input(self) -> None:
        """Toggle the visibility of the cell input."""
        if self.json["metadata"].get("jupyter", {}).get("source_hidden", False):
            self.show_input()
        else:
            self.hide_input()

    def show_output(self) -> None:
        """Set the cell outputs to visible."""
        self.set_metadata(("jupyter", "outputs_hidden"), False)
        self.set_metadata(("collapsed",), False)

    def hide_output(self) -> None:
        """Set the cell outputs to visible."""
        self.set_metadata(("jupyter", "outputs_hidden"), True)
        self.set_metadata(("collapsed",), True)

    def toggle_output(self) -> None:
        """Toggle the visibility of the cell outputs."""
        if self.json["metadata"].get("jupyter", {}).get("outputs_hidden", False):
            self.show_output()
        else:
            self.hide_output()

    def set_metadata(self, path: tuple[str, ...], data: Any) -> None:
        """Set a value in the metadata at an arbitrary path.

        Args:
            path: A tuple of path level names to create
            data: The value to add

        """
        level = self.json["metadata"]
        for i, key in enumerate(path):
            if i == len(path) - 1:
                level[key] = data
            else:
                level = level.setdefault(key, {})

    def set_status(self, status: str) -> None:
        """Set the execution status of the cell."""
        # log.debug(status)
        pass

    def get_input(
        self,
        prompt: str = "Please enter a value:",
        password: bool = False,
    ) -> None:
        """Scroll the cell requesting input into view and render it before asking for input."""
        self.kernel_tab.select(self.index)
        self.stdin_box.get_input(prompt, password)

    async def edit_in_editor(self) -> None:
        """Edit the cell in $EDITOR."""
        buffer = self.input_box.buffer
        app = get_app()
        edit_in_fg = False

        # Save VISUAL environment variable
        visual = os.environ.get("VISUAL")

        if editor := app.config.external_editor:
            if "{left}" in editor:
                win = self.input_box.window

                if (info := win.render_info) is not None:
                    edit_in_fg = True

                    margin_left = sum(
                        [win._get_margin_width(m) for m in win.left_margins]
                    )
                    margin_right = sum(
                        [win._get_margin_width(m) for m in win.right_margins]
                    )
                    top = info._y_offset
                    left = info._x_offset - margin_left
                    width = info.window_width + margin_left + margin_right
                    height = min(app.output.get_size().rows, info.window_height)

                else:
                    left = top = 0
                    height, width = app.output.get_size()

                editor = editor.format(
                    top=top,
                    left=left,
                    width=width,
                    height=height,
                    bottom=top + height,
                    right=left + width,
                )

            # Override VISUAL environment variable
            os.environ["VISUAL"] = editor

        if edit_in_fg:
            # Create a tempfile
            if buffer.tempfile:
                filename, cleanup_func = buffer._editor_complex_tempfile()
            else:
                filename, cleanup_func = buffer._editor_simple_tempfile()
            try:
                # Edit the temp file
                success = buffer._open_file_in_editor(filename)
                # Read content again.
                if success:
                    with open(filename, "rb") as f:
                        text = f.read().decode("utf-8")
                        # Drop trailing newline
                        if text.endswith("\n"):
                            text = text[:-1]
                        buffer.document = Document(text=text, cursor_position=len(text))
                    # Run the cell if configured
                    if app.config.run_after_external_edit:
                        buffer.validate_and_handle()
            finally:
                # Clean up temp dir/file.
                cleanup_func()

        else:
            await buffer.open_in_editor(
                validate_and_handle=app.config.run_after_external_edit
            )

        # Restore VISUAL environment variable
        if visual is not None:
            os.environ["VISUAL"] = visual

    # ################################### Settings ####################################

    add_setting(
        name="show_cell_borders",
        title="cell borders",
        flags=["--show-cell-borders"],
        type_=bool,
        help_="Show or hide cell borders.",
        default=False,
        schema={
            "type": "boolean",
        },
        description="""
            Whether cell borders should be drawn for unselected cells.
        """,
    )

    add_setting(
        name="external_editor",
        flags=["--external-editor"],
        type_=str,
        help_="Set the external editor to use.",
        default=None,
        description="""
            A command to run when editing cells externally. The following strings in
            the command will be replaced with values which locate the cell being
            edited:

            * ``{top}``
            * ``{left}``
            * ``{bottom}``
            * ``{right}``
            * ``{width}``
            * ``{height}``

            This is useful if you run euporie inside a tmux session, and wish to launch
            your editor in a pop-up pane. This can be achieved by setting this parameter
            to something like the following:

            .. code-block::

               "tmux display-popup -x {left} -y {bottom} -w {width} -h {height} -B -E micro"

        """,
    )



"""Define a command object for use in key-bindings, menus, and the command palette."""

from __future__ import annotations

import logging
import weakref
from inspect import isawaitable, iscoroutinefunction, signature
from typing import TYPE_CHECKING, cast

from prompt_toolkit.application import get_app
from prompt_toolkit.filters import to_filter
from prompt_toolkit.key_binding.key_bindings import Binding
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from euporie.core.key_binding.utils import parse_keys
from euporie.core.keys import Keys

if TYPE_CHECKING:
    from typing import Any, Callable, Coroutine

    from prompt_toolkit.filters import Filter, FilterOrBool
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        KeyHandlerCallable,
        NotImplementedOrNone,
    )

    from euporie.core.widgets.menu import MenuItem

    AnyKey = tuple[Keys | str, ...] | Keys | str
    AnyKeys = list[AnyKey] | AnyKey
    CommandHandlerNoArgs = Callable[
        ..., Coroutine[Any, Any, None] | NotImplementedOrNone
    ]
    CommandHandlerArgs = Callable[
        [KeyPressEvent], Coroutine[Any, Any, None] | NotImplementedOrNone
    ]
    CommandHandler = CommandHandlerNoArgs | CommandHandlerArgs

log = logging.getLogger(__name__)


class Command:
    """Wrap a function so it can be used as a key-binding or a menu item."""

    def __init__(
        self,
        handler: CommandHandler,
        *,
        filter: FilterOrBool = True,
        hidden: FilterOrBool = False,
        name: str | None = None,
        title: str | None = None,
        menu_title: str | None = None,
        description: str | None = None,
        toggled: Filter | None = None,
        eager: FilterOrBool = False,
        is_global: FilterOrBool = False,
        save_before: Callable[[KeyPressEvent], bool] = (lambda event: True),
        record_in_macro: FilterOrBool = True,
    ) -> None:
        """Create a new instance of a command.

        Args:
            handler: The callable to run when the command is triggers
            filter: The condition under which the command is allowed to run
            hidden: The condition under the command is visible to the user
            name: The name of the command, for accessing the command from the registry
            title: The title of the command for display
            menu_title: The title to display in menus if different
            description: The description of the command to explain it's function
            toggled: The toggle state of this command If this command toggles something
            eager: When True, ignore potential longer matches for this key binding
            is_global: Make this a global (always active) binding
            save_before: Determines if the buffer should be saved before running
            record_in_macro: Whether these key bindings should be recorded in macros

        """
        self.handler = handler
        self.filter = to_filter(filter)
        self.hidden = to_filter(hidden)
        if name is None:
            name = handler.__name__.strip("_").replace("_", "-")
        self.name = name
        if title is None:
            title = name.capitalize().replace("-", " ")
        self.title = title
        self.menu_title = menu_title or title
        if description is None:
            # Use the first line of the docstring as the command description
            if handler.__doc__:
                description = (
                    "".join(handler.__doc__.strip().split("\n")).split(".")[0] + "."
                )
            else:
                description = title or name.capitalize()
        self.description = description

        self.toggled = toggled
        self._menu: MenuItem | None = None

        self.eager = to_filter(eager)
        self.is_global = to_filter(is_global)
        self.save_before = save_before
        self.record_in_macro = to_filter(record_in_macro)

        self.keys: list[tuple[str | Keys, ...]] = []

    def run(self) -> None:
        """Run the command's handler."""
        if self.filter():
            app = get_app()
            result = self.key_handler(
                KeyPressEvent(
                    key_processor_ref=weakref.ref(app.key_processor),
                    arg=None,
                    key_sequence=[],
                    previous_key_sequence=[],
                    is_repeat=False,
                )
            )
            if isawaitable(result):

                async def _wait_and_invalidate() -> None:
                    assert isawaitable(result)
                    output = await result
                    if output is None:
                        app.invalidate()

                app.create_background_task(_wait_and_invalidate())
            elif result is None:
                app.invalidate()

    @property
    def key_handler(self) -> KeyHandlerCallable:
        """Return a key handler for the command."""
        handler = self.handler
        sig = signature(handler)

        if sig.parameters:
            # The handler already accepts a `KeyPressEvent` argument
            return cast("KeyHandlerCallable", handler)

        if iscoroutinefunction(handler):

            async def _key_handler_async(event: KeyPressEvent) -> NotImplementedOrNone:
                result = cast("CommandHandlerNoArgs", handler)()
                assert isawaitable(result)
                return await result

            return _key_handler_async

        else:

            def _key_handler(event: KeyPressEvent) -> NotImplementedOrNone:
                return cast("CommandHandlerNoArgs", handler)()

            return _key_handler

    def bind(self, key_bindings: KeyBindingsBase, keys: AnyKeys) -> None:
        """Add the current commands to a set of key bindings.

        Args:
            key_bindings: The set of key bindings to bind to
            keys: Additional keys to bind to the command

        """
        for key in parse_keys(keys):
            self.keys.append(key)
            key_bindings.bindings.append(
                Binding(
                    key,
                    handler=self.key_handler,
                    filter=self.filter,
                    eager=self.eager,
                    is_global=self.is_global,
                    save_before=self.save_before,
                    record_in_macro=self.record_in_macro,
                )
            )

    def key_str(self) -> str:
        """Return a string representing the first registered key-binding."""
        from euporie.core.key_binding.utils import format_keys

        if self.keys:
            return format_keys([self.keys[0]])[0]
        return ""

    @property
    def menu_handler(self) -> Callable[[], None]:
        """Return a menu handler for the command."""
        handler = self.handler
        if isawaitable(handler):

            def _menu_handler() -> None:
                task = cast("CommandHandlerNoArgs", handler)()
                task = cast("Coroutine[Any, Any, None]", task)
                if task is not None:
                    get_app().create_background_task(task)

            return _menu_handler
        else:
            return cast("Callable[[], None]", handler)

    @property
    def menu(self) -> MenuItem:
        """Return a menu item for the command."""
        from euporie.core.widgets.menu import MenuItem

        if self._menu is None:
            self._menu = MenuItem.from_command(self)
        return self._menu


commands: dict[str, Command] = {}


def add_cmd(**kwargs: Any) -> Callable:
    """Add a command to the centralized command system."""

    def decorator(handler: Callable) -> Callable:
        cmd = Command(handler, **kwargs)
        commands[cmd.name] = cmd
        return handler

    return decorator


def get_cmd(name: str) -> Command:
    """Get a command from the centralized command system by name.

    Args:
        name: The name of the command to retrieve

    Returns:
        The requested command object

    Raises:
        KeyError: Raised if the named command is not found

    """
    try:
        return commands[name]
    except KeyError as e:
        raise KeyError("Unknown command: %r" % name) from e



"""Contain functions which convert data to png format."""

from __future__ import annotations

import asyncio
from functools import partial
from typing import TYPE_CHECKING

from euporie.core.convert.core import register
from euporie.core.convert.formats.common import base64_to_bytes_py, imagemagick_convert
from euporie.core.convert.utils import commands_exist, have_modules

if TYPE_CHECKING:
    from pathlib import Path

    from PIL.Image import Image as PilImage


register(
    from_="base64-png",
    to="png",
)(base64_to_bytes_py)


@register(
    from_="latex",
    to="png",
    filter_=commands_exist("dvipng") & commands_exist("latex"),
)
async def latex_to_png_dvipng(
    data: str,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
    timeout: int = 2,
) -> bytes | None:
    """Render LaTeX as a png image using :command:`dvipng`.

    Borrowed from IPython.
    """
    import shutil
    import subprocess
    import tempfile
    from pathlib import Path

    latex_doc = (
        r"\documentclass{article}\pagestyle{empty}\begin{document}"
        + data
        + r"\end{document}"
    )

    workdir = Path(tempfile.mkdtemp())
    with workdir.joinpath("tmp.tex").open("w", encoding="utf8") as f:
        f.writelines(latex_doc)

    # Convert hex color to latex color
    if fg and len(fg) == 4:
        fg = f"#{fg[1]}{fg[1]}{fg[2]}{fg[2]}{fg[3]}{fg[3]}"
    fg_latex = (
        f"RGB {int(fg[1:3], 16)} {int(fg[3:5], 16)} {int(fg[5:7], 16)}" if fg else ""
    )

    # Convert latex document to dvi image, then Convert dvi image to png
    try:
        proc = await asyncio.create_subprocess_exec(
            *["latex", "-halt-on-error", "-interaction", "batchmode", "tmp.tex"],
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
            cwd=workdir,
        )
        await asyncio.wait_for(proc.wait(), timeout)

        dvipng_cmd = [
            "dvipng",
            "-T",
            "tight",
            "-D",
            "150",
            "-z",
            "9",
            "-bg",
            "Transparent",
            "-o",
            "/dev/stdout",
            "tmp.dvi",
        ]
        if fg:
            dvipng_cmd += ["-fg", fg_latex]

        proc = await asyncio.create_subprocess_exec(
            *dvipng_cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.DEVNULL,
            cwd=workdir,
        )
        output, _ = await asyncio.wait_for(proc.communicate(), timeout)
    except (subprocess.CalledProcessError, TimeoutError):
        return None
    finally:
        # Clean up temporary folder
        shutil.rmtree(workdir)

    return output


@register(
    from_="latex",
    to="png",
    filter_=have_modules("matplotlib"),
)
async def latex_to_png_py_mpl(
    data: str,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> bytes:
    """Render LaTeX as a png image using :py:module:`matplotlib`.

    Borrowed from IPython.
    """
    from io import BytesIO

    from matplotlib import figure, font_manager, mathtext
    from matplotlib.backends import backend_agg

    # mpl mathtext doesn't support display math, force inline
    data = data.replace("$$", "$")

    buffer = BytesIO()
    prop = font_manager.FontProperties(size=12)
    parser = mathtext.MathTextParser("path")
    width, height, depth, _, _ = parser.parse(data, dpi=72, prop=prop)
    fig = figure.Figure(figsize=(width or 256 / 72, height or 256 / 72))
    fig.text(0, depth / height, data, fontproperties=prop, color=fg)
    backend_agg.FigureCanvasAgg(fig)
    fig.savefig(buffer, dpi=120, format="png", transparent=True)
    return buffer.getvalue()


register(
    from_=("svg", "jpeg", "pdf", "gif"),
    to="png",
    filter_=commands_exist("convert", "magick"),
)(partial(imagemagick_convert, "PNG"))


@register(
    from_="pil",
    to="png",
    filter_=have_modules("PIL"),
)
async def pil_to_png_py_pil(
    data: PilImage,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> bytes:
    """Convert a pillow image to sixels :py:mod:`teimpy`."""
    import io

    with io.BytesIO() as output:
        data.save(output, format="PNG")
        contents = output.getvalue()
    return contents


@register(
    from_="svg",
    to="png",
    filter_=have_modules("cairosvg"),
)
async def svg_to_png_py_cairosvg(
    data: str | bytes,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> str:
    """Convert SVG to PNG using :py:mod:`cairosvg`."""
    import cairosvg

    markup = data.decode() if isinstance(data, bytes) else data
    return cairosvg.surface.PNGSurface.convert(markup, write_to=None)



"""Define a file browser widget."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.key_binding.key_bindings import KeyBindings, KeyBindingsBase
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.layout.screen import WritePosition
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core.app import get_app
from euporie.core.border import InsetGrid
from euporie.core.config import add_setting
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.widgets.decor import Border, FocusedStyle
from euporie.core.widgets.forms import Button, Text

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Callable

    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from upath.core import PT

    from euporie.core.widgets.status_bar import StatusBarFields

log = logging.getLogger(__name__)


FILE_ICONS = {
    # default
    "file": "",
    "dir": "",
    # file extensions (vim-devicons)
    ".styl": "",
    ".sass": "",
    ".scss": "",
    ".htm": "",
    ".html": "",
    ".slim": "",
    ".haml": "",
    ".ejs": "",
    ".css": "",
    ".less": "",
    ".md": "",
    ".mdx": "",
    ".markdown": "",
    ".rmd": "",
    ".json": "",
    ".webmanifest": "",
    ".js": "",
    ".mjs": "",
    ".jsx": "",
    ".rb": "",
    ".gemspec": "",
    ".rake": "",
    ".php": "",
    ".py": "",
    ".pyc": "",
    ".pyo": "",
    ".pyd": "",
    ".coffee": "",
    ".mustache": "",
    ".hbs": "",
    ".conf": "",
    ".ini": "",
    ".yml": "",
    ".yaml": "",
    ".toml": "",
    ".bat": "",
    ".mk": "",
    ".twig": "",
    ".cpp": "",
    ".c++": "",
    ".cxx": "",
    ".cc": "",
    ".cp": "",
    ".c": "",
    ".cs": "",
    ".h": "",
    ".hh": "",
    ".hpp": "",
    ".hxx": "",
    ".hs": "",
    ".lhs": "",
    ".nix": "",
    ".lua": "",
    ".java": "",
    ".sh": "",
    ".fish": "",
    ".bash": "",
    ".zsh": "",
    ".ksh": "",
    ".csh": "",
    ".awk": "",
    ".ps1": "",
    ".ml": "λ",
    ".mli": "λ",
    ".diff": "",
    ".db": "",
    ".sql": "",
    ".dump": "",
    ".clj": "",
    ".cljc": "",
    ".cljs": "",
    ".edn": "",
    ".scala": "",
    ".go": "",
    ".dart": "",
    ".xul": "",
    ".sln": "",
    ".suo": "",
    ".pl": "",
    ".pm": "",
    ".t": "",
    ".rss": "",
    ".f#": "",
    ".fsscript": "",
    ".fsx": "",
    ".fs": "",
    ".fsi": "",
    ".rs": "",
    ".rlib": "",
    ".d": "",
    ".erl": "",
    ".hrl": "",
    ".ex": "",
    ".exs": "",
    ".eex": "",
    ".leex": "",
    ".heex": "",
    ".vim": "",
    ".ai": "",
    ".psd": "",
    ".psb": "",
    ".ts": "",
    ".tsx": "",
    ".jl": "",
    ".pp": "",
    ".vue": "﵂",
    ".elm": "",
    ".swift": "",
    ".xcplayground": "",
    ".tex": "ﭨ",
    ".r": "ﳒ",
    ".rproj": "鉶",
    ".sol": "ﲹ",
    ".pem": "",
    # file names (vim-devicons) (case-insensitive not supported in lf)
    "gruntfile.coffee": "",
    "gruntfile.js": "",
    "gruntfile.ls": "",
    "gulpfile.coffee": "",
    "gulpfile.js": "",
    "gulpfile.ls": "",
    "mix.lock": "",
    "dropbox": "",
    ".ds_store": "",
    ".gitconfig": "",
    ".gitignore": "",
    ".gitattributes": "",
    ".gitlab-ci.yml": "",
    ".bashrc": "",
    ".zshrc": "",
    ".zshenv": "",
    ".zprofile": "",
    ".vimrc": "",
    ".gvimrc": "",
    "_vimrc": "",
    "_gvimrc": "",
    ".bashprofile": "",
    "favicon.ico": "",
    "license": "",
    "node_modules": "",
    "react.jsx": "",
    "procfile": "",
    "dockerfile": "",
    "docker-compose.yml": "",
    "rakefile": "",
    "config.ru": "",
    "gemfile": "",
    "makefile": "",
    "cmakelists.txt": "",
    "robots.txt": "ﮧ",
    # file names (case-sensitive adaptations)
    "Gruntfile.coffee": "",
    "Gruntfile.js": "",
    "Gruntfile.ls": "",
    "Gulpfile.coffee": "",
    "Gulpfile.js": "",
    "Gulpfile.ls": "",
    "Dropbox": "",
    ".DS_Store": "",
    "LICENSE": "",
    "React.jsx": "",
    "Procfile": "",
    "Dockerfile": "",
    "Docker-compose.yml": "",
    "Rakefile": "",
    "Gemfile": "",
    "Makefile": "",
    "CMakeLists.txt": "",
    # file patterns (file name adaptations)
    "jquery.min.js": "",
    "angular.min.js": "",
    "backbone.min.js": "",
    "require.min.js": "",
    "materialize.min.js": "",
    "materialize.min.css": "",
    "mootools.min.js": "",
    "vimrc": "",
    "Vagrantfile": "",
    # archives or compressed (extensions from dircolors defaults)
    ".tar": "",
    ".tgz": "",
    ".arc": "",
    ".arj": "",
    ".taz": "",
    ".lha": "",
    ".lz4": "",
    ".lzh": "",
    ".lzma": "",
    ".tlz": "",
    ".txz": "",
    ".tzo": "",
    ".t7z": "",
    ".zip": "",
    ".z": "",
    ".dz": "",
    ".gz": "",
    ".lrz": "",
    ".lz": "",
    ".lzo": "",
    ".xz": "",
    ".zst": "",
    ".tzst": "",
    ".bz2": "",
    ".bz": "",
    ".tbz": "",
    ".tbz2": "",
    ".tz": "",
    ".deb": "",
    ".rpm": "",
    ".jar": "",
    ".war": "",
    ".ear": "",
    ".sar": "",
    ".rar": "",
    ".alz": "",
    ".ace": "",
    ".zoo": "",
    ".cpio": "",
    ".7z": "",
    ".rz": "",
    ".cab": "",
    ".wim": "",
    ".swm": "",
    ".dwm": "",
    ".esd": "",
    # image formats (extensions from dircolors defaults)
    ".jpg": "",
    ".jpeg": "",
    ".mjpg": "",
    ".mjpeg": "",
    ".webp": "",
    ".ico": "",
    ".gif": "",
    ".bmp": "",
    ".pbm": "",
    ".pgm": "",
    ".ppm": "",
    ".tga": "",
    ".xbm": "",
    ".xpm": "",
    ".tif": "",
    ".tiff": "",
    ".png": "",
    ".svg": "",
    ".svgz": "",
    ".mng": "",
    ".pcx": "",
    ".mov": "",
    ".mpg": "",
    ".mpeg": "",
    ".m2v": "",
    ".mkv": "",
    ".webm": "",
    ".ogm": "",
    ".mp4": "",
    ".m4v": "",
    ".mp4v": "",
    ".vob": "",
    ".qt": "",
    ".nuv": "",
    ".wmv": "",
    ".asf": "",
    ".rm": "",
    ".rmvb": "",
    ".flc": "",
    ".avi": "",
    ".fli": "",
    ".flv": "",
    ".gl": "",
    ".dl": "",
    ".xcf": "",
    ".xwd": "",
    ".yuv": "",
    ".cgm": "",
    ".emf": "",
    ".ogv": "",
    ".ogx": "",
    # audio formats (extensions from dircolors defaults)
    ".aac": "",
    ".au": "",
    ".flac": "",
    ".m4a": "",
    ".mid": "",
    ".midi": "",
    ".mka": "",
    ".mp3": "",
    ".mpc": "",
    ".ogg": "",
    ".ra": "",
    ".wav": "",
    ".oga": "",
    ".opus": "",
    ".spx": "",
    ".xspf": "",
    # documents
    ".pdf": "",
    ".doc": "",
    ".docx": "",
    ".ipynb": "ﴬ",
}


def is_dir(path: str | Path) -> bool | None:
    """Check if a path is a directory."""
    test_path = UPath(path)
    try:
        return test_path.is_dir()
    except (ValueError, PermissionError, TypeError):
        return None


class FileBrowserControl(UIControl):
    """A control for browsing a filesystem."""

    def __init__(
        self,
        path: Path | None = None,
        on_chdir: Callable[[FileBrowserControl], None] | None = None,
        on_select: Callable[[FileBrowserControl], None] | None = None,
        on_open: Callable[[FileBrowserControl], None] | None = None,
        window: Window | None = None,
    ) -> None:
        """Initialize a new file browser instance."""
        self.dir = path or UPath(".")
        self.hovered: int | None = None
        self.selected: int | None = None
        self._dir_cache: FastDictCache[
            tuple[Path], list[tuple[bool, Path]]
        ] = FastDictCache(get_value=self.load_path, size=1)
        self.on_select = Event(self, on_select)
        self.on_chdir = Event(self, on_chdir)
        self.on_open = Event(self, on_open)

        self.window = window

        self.on_chdir.fire()

        self.key_bindings = kb = KeyBindings()

        @kb.add("up")
        @kb.add("<scroll-up>")
        def _move_up(event: KeyPressEvent) -> None:
            self.move_cursor_up()

        @kb.add("down")
        @kb.add("<scroll-down>")
        def _move_down(event: KeyPressEvent) -> None:
            self.move_cursor_down()

        @kb.add("home")
        def _home(event: KeyPressEvent) -> None:
            self.select(0)

        @kb.add("end")
        def _end(event: KeyPressEvent) -> None:
            self.select(len(self.contents) - 1)

        @kb.add("left")
        def _up(event: KeyPressEvent) -> None:
            self.dir = self.dir.parent

        @kb.add("space")
        @kb.add("enter")
        @kb.add("right")
        def _open(event: KeyPressEvent) -> None:
            return self.open_path()

    @property
    def contents(self) -> list[tuple[bool, Path]]:
        """Return the contents of the current folder."""
        return self._dir_cache[(self.dir,)]

    @property
    def dir(self) -> Path:
        """Return the current folder path."""
        return self._dir

    @dir.setter
    def dir(self, value: PT) -> None:
        """Set the current folder path."""
        dir_path = UPath(value)
        try:
            dir_path = dir_path.resolve()
        except NotImplementedError:
            pass
        if is_dir(dir_path):
            self._dir = dir_path
        else:
            log.warning("'%s' is not a directory, not changing directory", value)

    @property
    def path(self) -> Path:
        """Return the current selected path."""
        return self.contents[self.selected or 0][1]

    @staticmethod
    def load_path(path: Path) -> list[tuple[bool, Path]]:
        """Return the contents of a folder."""
        paths = [] if path.parent == path else [path / ".."]
        try:
            paths += list(path.iterdir())
        except PermissionError:
            pass
        is_dirs = []
        for child in paths:
            child_is_dir = is_dir(child)
            if child_is_dir is None:
                child_is_dir = True
            is_dirs.append(child_is_dir)
        return sorted(zip(is_dirs, paths), key=lambda x: (not x[0], x[1].name))

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the content for this user control."""
        paths = self.contents

        def get_line(i: int) -> StyleAndTextTuples:
            if i > len(paths) - 1:
                return []
            is_dir, child = paths[i]

            row = child.name

            if get_app().config.show_file_icons:
                icon = (
                    FILE_ICONS["dir"]
                    if is_dir
                    else FILE_ICONS.get(child.suffix)
                    or FILE_ICONS.get(child.name)
                    or FILE_ICONS["file"]
                )
                row = f"{icon} {row}"

            style = "class:row"
            if i % 2:
                style += " class:alt-row"
            if i == self.hovered:
                style += " class:hovered"
            if i == self.selected:
                style += " class:selection"

            return [(style, f" {row} ".ljust(width))]

        return UIContent(
            get_line=get_line,
            line_count=len(paths),
            cursor_position=Point(0, self.selected or 0),
            show_cursor=False,
        )

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events."""
        row = mouse_event.position.y
        app = get_app()
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_DOWN
        ):
            app.layout.current_control = self
            app.mouse_limits = None
            self.hovered = None
            return self.select(row, open_file=True)
        elif mouse_event.event_type == MouseEventType.MOUSE_MOVE:
            # Mark item as hovered if mouse is over the control
            if (
                self.window is not None
                and (info := self.window.render_info) is not None
            ):
                rowcol_to_yx = info._rowcol_to_yx
                abs_mouse_pos = (
                    mouse_event.position.x + info._x_offset,
                    mouse_event.position.y + info._y_offset - info.vertical_scroll,
                )
                if abs_mouse_pos == app.mouse_position:
                    row_col_vals = rowcol_to_yx.values()
                    y_min, x_min = min(row_col_vals)
                    y_max, x_max = max(row_col_vals)
                    app.mouse_limits = WritePosition(
                        xpos=x_min,
                        ypos=y_min,
                        width=x_max - x_min + 1,
                        height=y_max - y_min + 1,
                    )
                    return self.hover(row)
                else:
                    # Clear mouse limits if mouse is outside control
                    app.mouse_limits = None
                    self.hovered = None
                    return None

        return NotImplemented

    def select(self, row: int | None, open_file: bool = False) -> NotImplementedOrNone:
        """Select a file in the browser."""
        if row is None:
            row = 0
        row = min(max(0, row), len(self.contents) - 1)
        if self.selected != row:
            self.selected = row
            self.on_select.fire()
        elif open_file:
            self.open_path()
        return None

    def hover(self, row: int | None) -> NotImplementedOrNone:
        """Hover a file in the browser."""
        if row is not None:
            row = min(max(0, row), len(self.contents) - 1)
            if self.hovered != row:
                self.hovered = row
                return None
        return NotImplemented

    def open_path(self) -> None:
        """Open the selected file."""
        if self.selected is not None:
            is_dir, path = self.contents[self.selected]
            if is_dir:
                self.dir = path.resolve()
                self.hover(self.hovered)
                self.selected = None
                self.on_chdir.fire()
            else:
                self.on_open.fire()

    def move_cursor_down(self) -> None:
        """Request to move the cursor down."""
        index = self.selected
        if index is None:
            index = 0
        else:
            index += 1
        self.select(index)

    def move_cursor_up(self) -> None:
        """Request to move the cursor up."""
        index = self.selected
        if index is None:
            index = len(self.contents)
        else:
            index -= 1
        self.select(index)

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Key bindings specific to this user control."""
        return self.key_bindings

    def is_focusable(self) -> bool:
        """Determine that the file_browser is focusable."""
        return True

    def __pt_status__(self) -> StatusBarFields:
        """Show the selected or hovered path in the statusbar."""
        if self.hovered is not None:
            return [[("", str(self.contents[self.hovered][1]))]], []
        elif self.selected is not None:
            return [[("", str(self.contents[self.selected][1]))]], []
        return [], []


class FileBrowser:
    """A file browser."""

    completer = PathCompleter(only_directories=True)

    def __init__(
        self,
        path: Path | None = None,
        on_select: Callable[[Path], None] | None = None,
        on_open: Callable[[Path], None] | None = None,
        on_chdir: Callable[[Path], None] | None = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
        style: str = "",
        show_address_bar: FilterOrBool = True,
    ) -> None:
        """Create a new instance."""

        def _accept_path(buffer: Buffer) -> bool:
            control.dir = Path(buffer.text)
            return True

        def _validate_path(path: str) -> bool:
            return is_dir(path) or False

        text = Text(
            validation=_validate_path,
            accept_handler=_accept_path,
            completer=self.completer,
        )
        self.control = control = FileBrowserControl(
            path=path,
            on_chdir=lambda x: setattr(text, "text", str(x.dir)),
        )
        if on_select is not None:
            control.on_select += (
                lambda x: on_select(x.path) if callable(on_select) else None
            )
        if on_chdir is not None:
            control.on_chdir += (
                lambda x: on_chdir(x.path) if callable(on_chdir) else None
            )
        if on_open is not None:
            control.on_open += lambda x: on_open(x.path) if callable(on_open) else None

        self.container = HSplit(
            [
                ConditionalContainer(
                    VSplit(
                        [
                            FocusedStyle(text),
                            FocusedStyle(
                                Button(
                                    "Go",
                                    on_click=lambda x: setattr(
                                        control, "dir", text.text
                                    ),
                                )
                            ),
                        ]
                    ),
                    filter=to_filter(show_address_bar),
                ),
                FocusedStyle(
                    Border(
                        VSplit(
                            [
                                window := Window(
                                    control,
                                    style="class:face",
                                ),
                                MarginContainer(ScrollbarMargin(), target=window),
                            ]
                        ),
                        border=InsetGrid,
                        style="class:input,inset,border",
                    ),
                    style_hover="",
                ),
            ],
            style="class:file-browser " + style,
            width=width,
            height=height,
        )
        # Set control's window so it can determine it's position for mouse-over
        control.window = window

    def __pt_container__(self) -> AnyContainer:
        """Return the tree-view container's content."""
        return self.container

    def __pt_status__(self) -> StatusBarFields:
        """Show the selected or hovered path in the statusbar."""
        return self.control.__pt_status__()

    # ################################### Settings ####################################

    add_setting(
        name="show_file_icons",
        flags=["--show-file-icons"],
        type_=bool,
        title="File icons",
        help_="Show file icons in the file manager",
        default=False,
        schema={
            "type": "boolean",
        },
        description="""
            Whether file icons should be shown in the file manager.

            These icons exist in the unicode private use area, and may require custom
            fonts such as ``awesome-terminal-fonts`` or ``nerdfonts`` to be installed.
        """,
    )



"""Contain functions which convert data to html format."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from markdown_it import MarkdownIt
from mdit_py_plugins.amsmath import amsmath_plugin
from mdit_py_plugins.dollarmath.index import dollarmath_plugin
from mdit_py_plugins.texmath.index import texmath_plugin
from pygments import highlight
from pygments.formatters import HtmlFormatter

from euporie.core.convert.core import register
from euporie.core.current import get_app
from euporie.core.lexers import detect_lexer

if TYPE_CHECKING:
    from pathlib import Path

log = logging.getLogger(__name__)


class MarkdownParser(MarkdownIt):
    """Subclas the markdown parser to allow ``file:`` URIs."""

    def validateLink(self, url: str) -> bool:
        """Allow all link URIs."""
        return True


markdown_parser = (
    (
        MarkdownParser(
            options_update={
                "highlight": lambda text, language, lang_args: highlight(
                    text,
                    detect_lexer(text, language=language),
                    HtmlFormatter(
                        nowrap=True,
                        noclasses=True,
                        style=app.config.syntax_theme
                        if hasattr((app := get_app()), "config")
                        else "default",
                    ),
                )
            }
        )
        .enable("linkify")
        .enable("table")
        .enable("strikethrough")
    )
    .use(texmath_plugin)
    .use(dollarmath_plugin)
    .use(amsmath_plugin)
    # .use(tasklists_plugin)
)


@register(from_="markdown", to="html")
async def markdown_to_html_markdown_it(
    data: str | bytes,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> str:
    """Convert markdown to HTML using :py:mod:`markdownit_py`."""
    assert markdown_parser is not None
    markup = data.decode() if isinstance(data, bytes) else data
    return markdown_parser.render(markup)



"""Tab for use in euporie notebook editor."""

from euporie.notebook.tabs.display import DisplayTab
from euporie.notebook.tabs.edit import EditorTab
from euporie.notebook.tabs.json import JsonTab
from euporie.notebook.tabs.log import LogView
from euporie.notebook.tabs.notebook import Notebook
from euporie.web.tabs.web import WebTab

__all__ = ["DisplayTab", "EditorTab", "JsonTab", "LogView", "Notebook", "WebTab"]



"""Dialog."""

from __future__ import annotations

import logging
import traceback
from abc import ABCMeta, abstractmethod
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.clipboard import ClipboardData
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import (
    Condition,
    buffer_has_focus,
    has_completions,
    has_focus,
)
from prompt_toolkit.formatted_text import AnyFormattedText, to_formatted_text
from prompt_toolkit.formatted_text.utils import split_lines
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.key_binding.key_bindings import DynamicKeyBindings, KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    Float,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl, UIContent, UIControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.screen import WritePosition
from prompt_toolkit.mouse_events import MouseButton, MouseEventType
from prompt_toolkit.widgets.base import Box, Label

from euporie.core.app import get_app
from euporie.core.border import (
    FullLine,
    LowerLeftHalfLine,
    UpperRightHalfLine,
)
from euporie.core.commands import add_cmd
from euporie.core.filters import tab_has_focus
from euporie.core.ft.utils import FormattedTextAlign, align, lex
from euporie.core.key_binding.registry import register_bindings
from euporie.core.tabs.base import Tab
from euporie.core.widgets.decor import Border, FocusedStyle, Shadow
from euporie.core.widgets.file_browser import FileBrowser
from euporie.core.widgets.forms import Button, LabelledWidget, Select, Text

if TYPE_CHECKING:
    from typing import Any, Callable, Hashable

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.layout import FocusableElement
    from prompt_toolkit.mouse_events import MouseEvent

    from euporie.core.app import BaseApp
    from euporie.core.tabs.base import KernelTab
    from euporie.core.widgets.file_browser import FileBrowserControl

log = logging.getLogger(__name__)


DialogGrid = (
    FullLine.left_edge
    + FullLine.right_edge
    + LowerLeftHalfLine.bottom_edge
    + UpperRightHalfLine.top_edge
)


class DialogTitleControl(UIControl):
    """A draggable dialog titlebar."""

    def __init__(self, title: AnyFormattedText, dialog: Dialog, window: Window) -> None:
        """Initialize a new dialog titlebar."""
        self.title = title
        self.dialog = dialog
        self.window = window
        self.drag_start: Point | None = None
        self._content_cache: SimpleCache[Hashable, UIContent] = SimpleCache(maxsize=18)

    def create_content(self, width: int, height: int | None) -> UIContent:
        """Create the title text content."""

        def get_content() -> UIContent:
            lines = list(
                split_lines(
                    align(
                        to_formatted_text(self.title), FormattedTextAlign.CENTER, width
                    )
                )
            )
            return UIContent(
                get_line=lambda i: lines[i],
                line_count=len(lines),
                show_cursor=False,
            )

        return self._content_cache.get((width,), get_content)

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Move the dialog when the titlebar is dragged."""
        if (info := self.window.render_info) is not None:
            # Get the global mouse position
            app = self.dialog.app
            gx, gy = app.mouse_position
            if mouse_event.button == MouseButton.LEFT:
                if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                    # Start the drag event
                    self.drag_start = mouse_event.position
                    # Send all mouse events to this position
                    y_min, x_min = min(info._rowcol_to_yx.values())
                    y_max, x_max = max(info._rowcol_to_yx.values())
                    app.mouse_limits = WritePosition(
                        xpos=x_min,
                        ypos=y_min,
                        width=x_max - x_min,
                        height=y_max - y_min,
                    )
                    return NotImplemented
                elif mouse_event.event_type == MouseEventType.MOUSE_MOVE:
                    if self.drag_start is not None:
                        # Get available space
                        max_y, max_x = app.output.get_size()
                        # Calculate dialog dimensions
                        dl_width = self.dialog.content.preferred_width(max_x).preferred
                        dl_height = self.dialog.content.preferred_height(
                            dl_width, max_y
                        ).preferred
                        # Calculate new dialog position
                        new_x = max(1, min(gx - self.drag_start.x, max_x - dl_width))
                        new_y = max(1, min(gy - self.drag_start.y, max_y - dl_height))
                        # Move dialog
                        self.dialog.left = new_x - 1
                        self.dialog.top = new_y - 1
                        # change the mouse capture position
                        if app.mouse_limits is not None:
                            app.mouse_limits.xpos = new_x
                            app.mouse_limits.ypos = new_y
                        return None

            # End the drag event
            self.drag_start = None
            # Stop capturing all mouse events
            app.mouse_limits = None
        return NotImplemented


class Dialog(Float, metaclass=ABCMeta):
    """A modal dialog which is displayed above the application.

    Returns focus to the previously selected control when closed.
    """

    title: AnyFormattedText | None = None
    body_padding_top = 1
    body_padding_bottom = 0

    def __init__(self, app: BaseApp) -> None:
        """Create a modal dialog.

        title: The title of the dialog. Can be formatted text.
        body: The container to use as the main body of the dialog.
        buttons: A dictionary mapping text to display as dialog buttons to
            callbacks to run when the button is clicked. If the callback is
            `None`, the dialog will be closed without running a callback.
        to_focus: The control to focus when the dialog is displayed.

        Args:
            app: The application the dialog is attached to

        """
        self.app = app
        self.to_focus: FocusableElement | None = None
        self.last_focused: FocusableElement | None = None
        self._visible = False
        self.visible = Condition(lambda: self._visible)

        # Set default body & buttons
        self.body: AnyContainer = Window()
        self.buttons: dict[str, Callable | None] = {"OK": None}
        self.button_widgets: list[AnyContainer] = []

        # Create key-bindings
        kb = KeyBindings()
        kb.add("escape")(lambda event: self.hide())
        kb.add("tab", filter=~has_completions)(focus_next)
        kb.add("s-tab", filter=~has_completions)(focus_previous)

        @kb.add("enter", filter=~has_completions & ~buffer_has_focus)
        def _focus_button(event: KeyPressEvent) -> NotImplementedOrNone:
            if self.button_widgets:
                app.layout.focus(self.button_widgets[0])
                return None
            return NotImplemented

        self.kb = kb
        self.buttons_kb = KeyBindings()

        # Create title row
        title_window = Window(height=1, style="class:dialog,dialog-title")
        title_window.content = DialogTitleControl(
            lambda: self.title, self, title_window
        )
        title_row = ConditionalContainer(
            title_window, filter=Condition(lambda: bool(self.title))
        )

        # Create body row with collapsible padding around the body.
        body_row = Box(
            body=DynamicContainer(lambda: self.body),
            padding=Dimension(preferred=1, max=1),
            padding_top=self.body_padding_top,
            padding_bottom=self.body_padding_bottom,
        )

        # The buttons.
        buttons_row = ConditionalContainer(
            Box(
                body=DynamicContainer(
                    lambda: VSplit(
                        self.button_widgets,
                        padding=1,
                        key_bindings=DynamicKeyBindings(lambda: self.buttons_kb),
                    )
                ),
                height=Dimension(min=1, max=3, preferred=3),
            ),
            filter=Condition(lambda: bool(self.buttons)),
        )

        # Create body
        self.container = ConditionalContainer(
            Shadow(
                Border(
                    HSplit(
                        [
                            title_row,
                            body_row,
                            buttons_row,
                        ],
                        style="class:dialog,body",
                        key_bindings=self.kb,
                        modal=False,
                    ),
                    border=DialogGrid,
                    style="class:dialog,border",
                ),
            ),
            filter=self.visible,
        )

        # Set the body as the float's contents
        super().__init__(content=self.container)

    def _button_handler(
        self, button: str = "", event: KeyPressEvent | None = None
    ) -> None:
        if callable(cb := self.buttons.get(button)):
            cb()
        else:
            self.hide()

    def _load(self, **params: Any) -> None:
        """Load body, create buttons, etc."""
        self.to_focus = None
        self.buttons_kb = KeyBindings()

        # Load body & buttons
        self.load(**params)

        # Create button widgets & callbacks
        self.button_widgets.clear()

        if self.buttons:
            width = max(map(len, self.buttons)) + 2
            used_keys = set()
            # Add each button
            for text in self.buttons:
                # Use the first letter as a key-binding if it's not already used
                if (key := text[0]) not in used_keys:
                    rest = text[1:]
                    used_keys |= {key}
                else:
                    key = ""
                    rest = text
                # Add a button with a handler
                handler = partial(self._button_handler, text)
                self.button_widgets.append(
                    FocusedStyle(
                        Button(
                            [("underline", key), ("", rest)],
                            on_click=handler,
                            width=width,
                            style="class:input",
                        )
                    )
                )
                # Add a key-handler
                if key:
                    self.buttons_kb.add(f"A-{key.lower()}", is_global=True)(handler)

        # When a button is selected, handle left/right key bindings.
        if len(self.button_widgets) > 1:
            first_selected = has_focus(self.button_widgets[0])
            last_selected = has_focus(self.button_widgets[-1])
            self.buttons_kb.add("left", filter=~first_selected)(focus_previous)
            self.buttons_kb.add("right", filter=~last_selected)(focus_next)

        # Focus first button by default if nothing else was specified by ``self.load()``
        if self.to_focus is None:
            self.to_focus = self.button_widgets[0]

    @abstractmethod
    def load(self) -> None:
        """Load the dialog's body etc."""

    def show(self, **params: Any) -> None:
        """Display and focuses the dialog."""
        # Reset position
        self.top = self.left = None
        # Re-draw the body
        self._load(**params)
        self.last_focused = self.app.layout.current_control
        self._visible = True
        if self.to_focus is not None:
            self.app.layout.focus(self.to_focus)
        else:
            try:
                self.app.layout.focus(self.container)
            except ValueError:
                pass
        self.app.layout.focus(self.container)
        self.app.invalidate()

    def hide(self, event: KeyPressEvent | None = None) -> None:
        """Hide the dialog."""
        self._visible = False
        if self.last_focused is not None:
            try:
                self.app.layout.focus(self.last_focused)
            except ValueError:
                self.app.layout.focus_next()
        # Stop any drag events
        self.app.mouse_limits = None

    def toggle(self) -> None:
        """Show or hides the dialog."""
        if self._visible:
            self.hide()
        else:
            self.show()

    def __pt_container__(self) -> AnyContainer:
        """Return the container's content."""
        return self.container


class AboutDialog(Dialog):
    """A dialog which shows an "about" message."""

    title = "About"

    def load(self) -> None:
        """Load the dialog's body."""
        from euporie.core import (
            __app_name__,
            __copyright__,
            __logo__,
            __strapline__,
            __version__,
        )

        self.body = Window(
            FormattedTextControl(
                [
                    ("class:logo", __logo__),
                    ("", " "),
                    ("bold", __app_name__),
                    ("", f"Version {__version__}\n\n".rjust(23, " ")),
                    ("", __strapline__.center(30)),
                    ("", "\n"),
                    ("class:hr", "─" * 30 + "\n\n"),
                    ("", __copyright__),
                    # ("", "\n"),
                ]
            ),
            dont_extend_height=True,
        )

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _about() -> None:
        """Show the about dialog."""
        from euporie.core.current import get_app

        if dialog := get_app().dialogs.get("about"):
            dialog.toggle()


class FileDialog(Dialog, metaclass=ABCMeta):
    """A base dialog to prompt the user for a file path."""

    title = "Select a File"

    def __init__(self, app: BaseApp) -> None:
        """Set up the dialog components on dialog initialization."""
        super().__init__(app)

        def _accept_text(buf: Buffer) -> bool:
            """Accept the text in the file input field and focuses the next field."""
            self.app.layout.focus_next()
            buf.complete_state = None
            return True

        completer = PathCompleter()
        self.filepath = Text(
            multiline=False,
            completer=completer,
            accept_handler=_accept_text,
            style="class:input",
            width=30,
        )
        self.file_browser = FileBrowser(
            height=Dimension(preferred=20),
            show_address_bar=False,
            on_select=lambda path: setattr(self.filepath, "text", path.name),
        )
        completer.get_paths = lambda: [str(self.file_browser.control.dir)]
        self.error = ""

        self.body = HSplit(
            [
                self.file_browser,
                FocusedStyle(LabelledWidget(self.filepath, "File name:")),
                ConditionalContainer(
                    Label(lambda: self.error, style="red"),
                    filter=Condition(lambda: bool(self.error)),
                ),
            ]
        )

    def load(
        self,
        text: str = "",
        tab: Tab | None = None,
        error: str = "",
        cb: Callable | None = None,
    ) -> None:
        """Load the dialog body."""
        self.filepath.text = text
        self.buttons = {
            "OK": partial(self.validate, self.filepath.buffer, tab=tab, cb=cb),
            "Cancel": None,
        }
        self.file_browser.control.on_open._handlers.clear()
        self.file_browser.control.on_open += lambda path: self.validate(
            self.filepath.buffer, tab=tab, cb=cb
        )
        self.file_browser.control.selected = None
        self.error = error
        self.to_focus = self.filepath

    def validate(
        self, buffer: Buffer, tab: Tab | None, cb: Callable | None = None
    ) -> None:
        """Validate the input."""
        self.hide()


class OpenFileDialog(FileDialog):
    """A dialog which prompts the user for a filepath to open."""

    title = "Select a File to Open"

    def __init__(self, app: BaseApp) -> None:
        """Additional body components."""
        from euporie.core.widgets.forms import Dropdown

        super().__init__(app)

        self.tab_dd = tab_dd = Dropdown(options=[])

        def _update_options(fb: FileBrowserControl) -> None:
            tabs = get_app().get_file_tabs(path) if (path := fb.path).is_file() else []
            tab_dd.options = tabs
            tab_dd.labels = [tab.name for tab in tabs]

        self.file_browser.control.on_select += _update_options

        if isinstance(self.body, HSplit):
            self.body.children.append(
                ConditionalContainer(
                    FocusedStyle(LabelledWidget(tab_dd, "Open with:")),
                    filter=Condition(lambda: len(tab_dd.options) > 1),
                )
            )

    def load(
        self,
        text: str = "",
        tab: Tab | None = None,
        error: str = "",
        cb: Callable | None = None,
    ) -> None:
        """Load the dialog body."""
        super().load(text=text, tab=tab, error=error, cb=cb)
        self.tab_dd.options = []
        self.tab_dd.labels = []

    def validate(
        self, buffer: Buffer, tab: Tab | None, cb: Callable | None = None
    ) -> None:
        """Validate the the file to open exists."""
        from upath import UPath

        from euporie.core.path import parse_path

        path = parse_path(self.file_browser.control.dir / buffer.text)
        if path is not None:
            if not path.exists():
                path = UPath(buffer.text)

            if path.exists():
                if path.is_dir():
                    self.file_browser.control.dir = path
                elif path.is_file():
                    self.hide()
                    self.app.open_file(path, tab_class=self.tab_dd.value)
                return
            else:
                self.show(
                    error="The file path specified does not exist", text=buffer.text
                )

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(menu_title="Open File…")
    def _open_file() -> None:
        """Open a file."""
        from euporie.core.current import get_app

        if dialog := get_app().dialogs.get("open-file"):
            dialog.show()

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.app.BaseApp": {
                "open-file": "c-o",
            }
        }
    )


class SaveAsDialog(FileDialog):
    """A dialog which prompts the user for a filepath to save the current tab."""

    title = "Select a Path to Save"

    def validate(
        self, buffer: Buffer, tab: Tab | None, cb: Callable | None = None
    ) -> None:
        """Validate the the file to open exists."""
        from euporie.core.path import parse_path

        path = parse_path(self.file_browser.control.dir / buffer.text)
        if tab and path is not None:
            if path.is_dir():
                self.file_browser.control.dir = path
            else:
                tab.save(path=path)
                self.hide()
                if callable(cb):
                    cb()

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(
        menu_title="Save As…",
        filter=tab_has_focus,
    )
    def _save_as() -> None:
        """Save the current file at a new location."""
        from euporie.core.current import get_app

        app = get_app()
        if dialog := app.dialogs.get("save-as"):
            dialog.show(tab=app.tab)

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.app.BaseApp": {
                "save-as": ("A-s"),
            }
        }
    )


class NoKernelsDialog(Dialog):
    """Dialog to warn the user that no installed kernels were found."""

    title = "No Kernels Found"

    def load(self) -> None:
        """Load the dialog body."""
        self.body = Window(
            FormattedTextControl(
                [
                    ("bold", "No Jupyter kernels were found.\n\n"),
                    ("", "You can view and edit the notebook,\n"),
                    ("", "but will not be able to run any code.\n\n"),
                    ("", "Try installing "),
                    ("class:md.code.inline", "ipykernel"),
                    ("", " by running:"),
                    ("", "\n\n"),
                    (
                        "class:md.code.inline",
                        "$ pip install --user ipykernel",
                    ),
                    ("", "\n"),
                ]
            )
        )


class SelectKernelDialog(Dialog):
    """A dialog which allows the user to select a kernel."""

    title = "Select Kernel"

    def load(
        self,
        kernel_specs: dict[str, Any] | None = None,
        runtime_dirs: dict[str, Path] | None = None,
        tab: KernelTab | None = None,
        message: str = "",
    ) -> None:
        """Load dialog body & buttons."""
        from jupyter_core.paths import jupyter_runtime_dir

        from euporie.core.widgets.layout import TabbedSplit

        kernel_specs = kernel_specs or {}
        runtime_dirs = runtime_dirs or {}

        options_specs = Select(
            options=list(kernel_specs.keys()),
            labels=[
                kernel_spec.get("spec", {}).get("display_name", kernel_name)
                for kernel_name, kernel_spec in kernel_specs.items()
            ],
            style="class:input,radio-buttons",
            prefix=("○", "◉"),
            multiple=False,
            border=None,
            rows=5,
            dont_extend_width=False,
        )
        self.to_focus = options_specs

        connection_files = {
            path.name: path
            for path in Path(jupyter_runtime_dir()).glob("kernel-*.json")
        }
        options_files = Select(
            options=list(connection_files.values()),
            labels=list(connection_files.keys()),
            style="class:input,radio-buttons",
            prefix=("○", "◉"),
            multiple=False,
            border=None,
            rows=5,
            dont_extend_width=False,
        )

        msg_ft = (f"{message}\n\n" if message else "") + "Please select a kernel:"

        self.body = HSplit(
            [
                Label(msg_ft),
                tabs := TabbedSplit(
                    [
                        FocusedStyle(options_specs),
                        FocusedStyle(options_files),
                    ],
                    titles=["New", "Existing"],
                    width=Dimension(min=30),
                ),
            ]
        )

        def _change_kernel() -> None:
            self.hide()
            assert tab is not None
            if tabs.active == 0:
                name = options_specs.value
                connection_file = None
            else:
                name = None
                connection_file = options_files.value
            tab.kernel.change(
                name=name, connection_file=connection_file, cb=tab.kernel_started
            )

        self.buttons = {
            "Select": _change_kernel,
            "Cancel": None,
        }


class MsgBoxDialog(Dialog):
    """A dialog which shows the user a message."""

    title = "Message"

    def load(self, title: str = "Message", message: str = "") -> None:
        """Load dialog body & buttons."""
        self.title = title
        self.body = Label(message)


class ConfirmDialog(Dialog):
    """A dialog which allows the user to confirm an action."""

    def load(
        self,
        title: str = "Are you sure?",
        message: str = "Please confirm",
        cb: Callable[[], None] | None = None,
    ) -> None:
        """Load dialog body & buttons."""
        self.title = title
        self.body = Label(message)

        def _callback() -> None:
            if callable(cb):
                cb()
            self.hide()

        self.buttons = {
            "Yes": _callback,
            "No": None,
        }


class ErrorDialog(Dialog):
    """A dialog to show unhandled exceptions."""

    title = "Error"

    def load(self, exception: Exception | None = None, when: str = "") -> None:
        """Load dialog body & buttons."""
        from euporie.core.margins import MarginContainer, ScrollbarMargin
        from euporie.core.widgets.formatted_text_area import FormattedTextArea
        from euporie.core.widgets.forms import Checkbox

        if exception is None:
            exception = Exception("Unspecified Error")

        checkbox = Checkbox(
            text="Show traceback",
            prefix=("▶", "▼"),
        )

        tb_text = "".join(
            traceback.format_exception(None, exception, exception.__traceback__)
        )

        self.body = HSplit(
            [
                Window(
                    FormattedTextControl(
                        [
                            ("bold", "An error occurred"),
                            ("bold", f" when {when}" if when else ""),
                            ("bold", ":"),
                            ("", "\n\n"),
                            ("fg:ansired", exception.__repr__()),
                            ("", "\n"),
                        ],
                    )
                ),
                FocusedStyle(
                    Box(checkbox, padding_left=0),
                ),
                ConditionalContainer(
                    VSplit(
                        [
                            fta := FormattedTextArea(
                                lex([("", tb_text)], "pytb"),
                                width=80,
                                height=Dimension(min=10),
                                wrap_lines=False,
                                style="",
                                focus_on_click=True,
                            ),
                            MarginContainer(ScrollbarMargin(), target=fta.window),
                        ]
                    ),
                    filter=Condition(lambda: checkbox.selected),
                ),
            ]
        )

        def _copy_traceback() -> None:
            self.app.clipboard.set_data(ClipboardData(tb_text))

        self.buttons = {"Close": None, "Copy Traceback": _copy_traceback}


class UnsavedDialog(Dialog):
    """A dialog prompting the user to save unsaved changes."""

    title = "Unsaved Changes"

    def load(
        self, tab: Tab | None = None, cb: Callable[[], None] | None = None
    ) -> None:
        """Load the dialog body."""
        tab = tab or self.app.tab
        assert tab is not None

        self.body = Window(
            FormattedTextControl(
                [
                    ("bold", tab.path.name if tab.path else tab.__class__.__name__),
                    ("", " has unsaved changes\n"),
                    ("", "Do you want to save your changes?"),
                ]
            ),
        )

        def yes_cb() -> None:
            assert tab is not None
            self.hide()
            tab.save(cb=partial(tab.close, cb))

        def no_cb() -> None:
            assert tab is not None
            self.hide()
            Tab.close(tab, cb)

        self.buttons = {
            "Yes": yes_cb,
            "No": no_cb,
            "Cancel": None,
        }


class ShortcutsDialog(Dialog):
    """Display details of registered key-bindings in a dialog."""

    title = "Keyboard Shortcuts"

    def __init__(self, app: BaseApp) -> None:
        """Create a new shortcuts dialog instance."""
        super().__init__(app)
        self.details: StyleAndTextTuples | None = None

    def load(self, *args: Any, **kwargs: Any) -> None:
        """Load the dialog body."""
        from euporie.core.ft.utils import max_line_width
        from euporie.core.margins import MarginContainer, ScrollbarMargin
        from euporie.core.widgets.formatted_text_area import FormattedTextArea

        if not self.details:
            self.details = self.format_key_info()
        assert self.details is not None

        width = max_line_width(self.details)

        fta = FormattedTextArea(
            formatted_text=self.details,
            multiline=True,
            focusable=True,
            wrap_lines=False,
            width=width - 2,
        )
        self.body = VSplit([fta, MarginContainer(ScrollbarMargin(), target=fta.window)])

    def format_key_info(self) -> StyleAndTextTuples:
        """Generate a table with the current key bindings."""
        import importlib
        from textwrap import dedent

        from prompt_toolkit.formatted_text.base import to_formatted_text

        from euporie.core.border import InvisibleLine
        from euporie.core.commands import get_cmd
        from euporie.core.data_structures import DiInt
        from euporie.core.ft.table import Table
        from euporie.core.ft.utils import FormattedTextAlign
        from euporie.core.key_binding.registry import BINDINGS
        from euporie.core.key_binding.utils import format_keys, parse_keys

        table = Table(padding=0)

        for group, bindings in BINDINGS.items():
            if any(not get_cmd(cmd_name).hidden() for cmd_name in bindings):
                mod_name, cls_name = group.rsplit(".", maxsplit=1)
                mod = importlib.import_module(mod_name)
                app_cls = getattr(mod, cls_name)
                section_title = (
                    dedent(app_cls.__doc__).strip().split("\n")[0].rstrip(".")
                )

                row = table.new_row()
                row.new_cell(
                    section_title,
                    align=FormattedTextAlign.CENTER,
                    colspan=2,
                    style="class:shortcuts.group",
                    border_visibility=True,
                    border_line=InvisibleLine,
                )
                for i, (cmd_name, keys) in enumerate(bindings.items()):
                    cmd = get_cmd(cmd_name)
                    if not cmd.hidden():
                        key_strs = format_keys(parse_keys(keys))
                        row = table.new_row(
                            style="class:shortcuts.row" + (",alt" if i % 2 else "")
                        )
                        row.new_cell(
                            "\n".join(key_strs),
                            align=FormattedTextAlign.RIGHT,
                            style="class:key",
                            border_visibility=False,
                            border_line=InvisibleLine,
                        )
                        row.new_cell(
                            cmd.title,
                            border_visibility=False,
                            border_line=InvisibleLine,
                        )
        table.padding = DiInt(0, 1, 0, 1)

        return to_formatted_text(table)

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _keyboard_shortcuts() -> None:
        """Display details of registered key-bindings in a dialog."""
        from euporie.core.current import get_app

        if dialog := get_app().dialogs.get("shortcuts"):
            dialog.toggle()



"""Define a status-bar widget."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from prompt_toolkit.cache import FastDictCache
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.formatted_text import to_formatted_text
from prompt_toolkit.layout import containers
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.containers import (
    to_container as ptk_to_container,
)
from prompt_toolkit.layout.controls import FormattedTextControl

from euporie.core.config import add_setting
from euporie.core.current import get_app
from euporie.core.filters import is_searching

if TYPE_CHECKING:
    from typing import Callable, Sequence

    from prompt_toolkit.filters.base import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.formatted_text.base import AnyFormattedText
    from prompt_toolkit.layout.containers import AnyContainer, Container

    StatusBarFields = tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]

log = logging.getLogger(__name__)


_CONTAINER_STATUSES: WeakKeyDictionary[
    Container, Callable[[], StatusBarFields | None]
] = WeakKeyDictionary()


def to_container(container: AnyContainer) -> Container:
    """Monkey-patch `to_container` to collect container status functions."""
    result = ptk_to_container(container)
    if hasattr(container, "__pt_status__"):
        _CONTAINER_STATUSES[result] = container.__pt_status__
    return result


containers.to_container = to_container


class StatusContainer:
    """A container which allows attaching a status function."""

    def __init__(
        self, body: AnyContainer, status: Callable[[], StatusBarFields | None]
    ) -> None:
        """Initiate a new instance with a body and status function."""
        self.body = body
        self.status = status

    def __pt_status__(self) -> StatusBarFields | None:
        """Return the status fields."""
        return self.status()

    def __pt_container__(self) -> AnyContainer:
        """Return the body container."""
        return self.body


class StatusBar:
    """A status bar which shows the status of the current tab."""

    def __init__(
        self, extra_filter: FilterOrBool = True, default: StatusBarFields | None = None
    ) -> None:
        """Create a new status bar instance."""
        self.default: StatusBarFields = default or ([], [])
        self._status_cache: FastDictCache[
            tuple[int], list[StyleAndTextTuples]
        ] = FastDictCache(self._status, size=1)
        self.container = ConditionalContainer(
            content=VSplit(
                [
                    Window(
                        FormattedTextControl(
                            lambda: self._status_cache[get_app().render_counter,][0]
                        ),
                        style="class:status",
                    ),
                    Window(
                        FormattedTextControl(
                            lambda: self._status_cache[get_app().render_counter,][1]
                        ),
                        style="class:status.right",
                        align=WindowAlign.RIGHT,
                    ),
                ],
                height=1,
            ),
            filter=get_app().config.filter("show_status_bar")
            & ~is_searching
            & to_filter(extra_filter),
        )

    def _status(self, render_counter: int = 0) -> list[StyleAndTextTuples]:
        """Load and format the current status bar entries."""
        layout = get_app().layout
        current: Container = layout.current_window

        entries = self.default
        while True:
            if callable(
                func := (
                    _CONTAINER_STATUSES.get(current)
                    or getattr(current, "__pt_status__", None)
                )
            ):
                result = func()
                if result is not None:
                    entries = result
                    break
            elif current in layout._child_to_parent:
                current = layout._child_to_parent[current]
                continue
            break

        output: list[StyleAndTextTuples] = []
        # Show the tab's status fields
        for entry in entries:
            output.append([])
            for field in entry:
                if field:
                    if isinstance(field, tuple):
                        ft = [field]
                    else:
                        ft = to_formatted_text(field, style="class:status.field")
                    output[-1] += [
                        ("class:status.field", " "),
                        *ft,
                        ("class:status.field", " "),
                        ("class:status", " "),
                    ]
            if output[-1]:
                output[-1].pop()
        return output

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container

    # ################################### Settings ####################################

    add_setting(
        name="show_status_bar",
        flags=["--show-status-bar"],
        type_=bool,
        title="status bar",
        help_="Show the status bar",
        default=True,
        schema={
            "type": "boolean",
        },
        description="""
            Whether the status bar should be shown at the bottom of the screen.
        """,
    )



"""Allow access to the current running application."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.application.current import _current_app_session

if TYPE_CHECKING:
    from euporie.core.app import BaseApp


def get_app() -> BaseApp:
    """Get the current active (running) Application."""
    from euporie.core.app import BaseApp

    session = _current_app_session.get()
    if isinstance(session.app, BaseApp):
        return session.app

    # Use a baseapp as our "DummyApplication"
    return BaseApp()



"""Define editor key-bindings and commands for the micro editing mode."""

from __future__ import annotations

import logging
import re
from functools import partial
from typing import TYPE_CHECKING

from aenum import extend_enum
from prompt_toolkit.buffer import indent, unindent
from prompt_toolkit.document import Document
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import (
    buffer_has_focus,
    has_selection,
    in_paste_mode,
    is_multiline,
    shift_selection_mode,
)
from prompt_toolkit.key_binding import ConditionalKeyBindings
from prompt_toolkit.key_binding.bindings.named_commands import (
    accept_line,
    backward_delete_char,
    backward_word,
    beginning_of_buffer,
    delete_char,
    end_of_buffer,
    forward_word,
    get_by_name,
)
from prompt_toolkit.key_binding.bindings.scroll import (
    scroll_backward,
    scroll_forward,
    scroll_half_page_down,
    scroll_half_page_up,
    scroll_one_line_down,
    scroll_one_line_up,
    scroll_page_down,
    scroll_page_up,
)
from prompt_toolkit.keys import Keys
from prompt_toolkit.selection import SelectionState, SelectionType

from euporie.core.commands import add_cmd, get_cmd
from euporie.core.current import get_app
from euporie.core.filters import (
    buffer_is_code,
    buffer_is_markdown,
    cursor_at_start_of_line,
    cursor_in_leading_ws,
    has_suggestion,
    insert_mode,
    is_returnable,
    micro_insert_mode,
    micro_mode,
    micro_recording_macro,
    micro_replace_mode,
)
from euporie.core.key_binding.micro_state import MicroInputMode
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.key_binding.utils import if_no_repeat

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindingsBase, KeyPressEvent

    from euporie.core.config import Config

log = logging.getLogger(__name__)


class EditMode:
    """Micro style editor key-bindings."""


# Register micro edit mode
extend_enum(EditingMode, "MICRO", "MICRO")

# Register default bindings for micro edit mode
register_bindings(
    {
        "euporie.core.key_binding.bindings.micro.EditMode": {
            "move-cursor-right": "right",
            "move-cursor-left": "left",
            "newline": "enter",
            "accept-line": "enter",
            "backspace": ["backspace", "c-h"],
            "backward-kill-word": [
                "c-backspace",
                "A-backspace",
                ("c-A-h"),
                ("escape", "c-h"),
            ],
            "start-selection": [
                "s-up",
                "s-down",
                "s-right",
                "s-left",
                ("A-s-left"),
                ("A-s-right"),
                "c-s-left",
                "c-s-right",
                "s-home",
                "s-end",
                "c-s-home",
                "c-s-end",
            ],
            "extend-selection": [
                "s-up",
                "s-down",
                "s-right",
                "s-left",
                ("A-s-left"),
                ("A-s-right"),
                "c-s-left",
                "c-s-right",
                "s-home",
                "s-end",
                "c-s-home",
                "c-s-end",
            ],
            "cancel-selection": [
                "up",
                "down",
                "right",
                "left",
                ("A-left"),
                ("A-right"),
                "c-left",
                "c-right",
                "home",
                "end",
                "c-home",
                "c-end",
            ],
            "replace-selection": "<any>",
            "delete-selection": ["delete", "backspace", "c-h"],
            "backward-word": ["c-left", ("A-b")],
            "forward-word": ["c-right", ("A-f")],
            "move-lines-up": ("A-up"),
            "move-lines-down": ("A-down"),
            "go-to-start-of-line": ["home", ("A-left"), ("A-a")],
            "go-to-end-of-line": ["end", ("A-right"), ("A-e")],
            "beginning-of-buffer": ["c-up", "c-home"],
            "end-of-buffer": ["c-down", "c-end"],
            "go-to-start-of-paragraph": ("A-{"),
            "go-to-end-of-paragraph": ("A-}"),
            "indent-lines": "tab",
            "unindent-line": "backspace",
            "unindent-lines": "s-tab",
            "undo": "c-z",
            "redo": "c-y",
            "copy-selection": "c-c",
            "cut-selection": ["c-x", "s-delete"],
            "cut-line": "c-k",
            "duplicate-line": "c-d",
            "duplicate-selection": "c-d",
            "paste-clipboard": "c-v",
            "select-all": "c-a",
            "scroll-page-up": "pageup",
            "scroll-page-down": "pagedown",
            "delete": "delete",
            "toggle-case": "f4",
            "toggle-overwrite-mode": "insert",
            "start-macro": "c-u",
            "end-macro": "c-u",
            "run-macro": "c-j",
            "accept-suggestion": ["right", "c-f"],
            "fill-sugestion": ("A-f"),
            "toggle-comment": "c-_",
            "go-to-matching-bracket": [("A-("), ("A-)")],
            'wrap-selection-""': '"',
            "wrap-selection-''": "'",
            "wrap-selection-()": ["(", ")"],
            "wrap-selection-{}": ["{", "}"],
            "wrap-selection-[]": ["[", "]"],
            "wrap-selection-``": "`",
            "wrap-selection-**": "*",
            "wrap-selection-__": "_",
        },
    }
)


def load_micro_bindings(config: Config | None = None) -> KeyBindingsBase:
    """Load editor key-bindings in the style of the ``micro`` text editor."""
    return ConditionalKeyBindings(
        load_registered_bindings(
            "euporie.core.key_binding.bindings.micro.EditMode", config=config
        ),
        micro_mode,
    )


# Commands


@add_cmd(
    filter=buffer_has_focus,
)
def toggle_overwrite_mode() -> None:
    """Toggle overwrite when using micro editing mode."""
    if micro_replace_mode():
        get_app().micro_state.input_mode = MicroInputMode.INSERT
    elif micro_insert_mode():
        get_app().micro_state.input_mode = MicroInputMode.REPLACE


@add_cmd(
    filter=buffer_has_focus & ~micro_recording_macro,
    hidden=True,
)
def start_macro() -> None:
    """Start recording a macro."""
    get_app().micro_state.start_macro()


@add_cmd(
    filter=buffer_has_focus & micro_recording_macro,
    hidden=True,
)
def end_macro() -> None:
    """Stop recording a macro."""
    get_app().micro_state.end_macro()


@add_cmd(
    filter=buffer_has_focus,
    record_in_macro=False,
    hidden=True,
)
def run_macro() -> None:
    """Re-execute the last keyboard macro defined."""
    # Insert the macro.
    app = get_app()
    macro = app.micro_state.macro
    if macro:
        app.key_processor.feed_multiple(macro, first=True)


add_cmd(
    name="backspace",
    title="Delete previous character",
    filter=buffer_has_focus & ~has_selection,
    save_before=if_no_repeat,
)(backward_delete_char)
add_cmd(
    title="Delete character", name="delete", filter=buffer_has_focus & ~has_selection
)(delete_char)


@add_cmd(title="Delete previous word", filter=buffer_has_focus)
def backward_kill_word(event: KeyPressEvent) -> None:
    """Delete the word behind the cursor, using whitespace as a word boundary."""
    buff = event.current_buffer
    pos = buff.document.find_start_of_previous_word()

    # Delete until the start of the document if nothing is found
    if pos is None:
        pos = -buff.cursor_position
    # Delete the word
    if pos:
        buff.delete_before_cursor(count=-pos)
    else:
        # Nothing to delete. Bell.
        event.app.output.bell()


# Navigation

add_cmd(
    title="Move back one word",
    filter=buffer_has_focus,
)(backward_word)
add_cmd(
    title="Move forward one word",
    filter=buffer_has_focus,
)(forward_word)
add_cmd(
    title="Move to the beginning of the input",
    filter=buffer_has_focus,
)(beginning_of_buffer)
add_cmd(
    title="Move to the end of the input",
    filter=buffer_has_focus,
)(end_of_buffer)

add_cmd(
    filter=buffer_has_focus,
)(scroll_backward)
add_cmd(
    filter=buffer_has_focus,
)(scroll_forward)
add_cmd(
    title="Scroll down half a page",
    filter=buffer_has_focus,
)(scroll_half_page_down)
add_cmd(
    title="Scroll up half a page",
    filter=buffer_has_focus,
)(scroll_half_page_up)
add_cmd(
    title="Scroll down one line",
    filter=buffer_has_focus,
)(scroll_one_line_down)
add_cmd(
    title="Scroll up one line",
    filter=buffer_has_focus,
)(scroll_one_line_up)
add_cmd(
    title="Scroll one one page",
    filter=buffer_has_focus,
)(scroll_page_down)
add_cmd(
    title="Scroll up one page",
    filter=buffer_has_focus,
)(scroll_page_up)


@add_cmd(
    filter=buffer_has_focus,
)
def move_cursor_left() -> None:
    """Move back a character, or up a line."""
    get_app().current_buffer.cursor_position -= 1


@add_cmd(
    filter=buffer_has_focus,
)
def move_cursor_right() -> None:
    """Move forward a character, or down a line."""
    get_app().current_buffer.cursor_position += 1


@add_cmd(
    filter=buffer_has_focus & ~shift_selection_mode,
)
def go_to_start_of_line() -> None:
    """Move the cursor to the start of the line."""
    buff = get_app().current_buffer
    buff.cursor_position += buff.document.get_start_of_line_position(
        after_whitespace=not cursor_in_leading_ws()
        or buff.document.cursor_position_col == 0
    )


@add_cmd(
    name="go-to-end-of-line",
    filter=buffer_has_focus & ~shift_selection_mode,
)
def go_to_end_of_line() -> None:
    """Move the cursor to the end of the line."""
    buff = get_app().current_buffer
    buff.cursor_position += buff.document.get_end_of_line_position()


@add_cmd(
    filter=buffer_has_focus,
)
def go_to_start_of_paragraph() -> None:
    """Move the cursor to the start of the current paragraph."""
    buf = get_app().current_buffer
    buf.cursor_position += buf.document.start_of_paragraph()


@add_cmd(
    filter=buffer_has_focus,
)
def go_to_end_of_paragraph() -> None:
    """Move the cursor to the end of the current paragraph."""
    buffer = get_app().current_buffer
    buffer.cursor_position += buffer.document.end_of_paragraph()


# Editing


@add_cmd(
    filter=buffer_has_focus & buffer_is_code,
)
def toggle_comment() -> None:
    """Comment or uncomments the current or selected lines."""
    comment = "# "
    buffer = get_app().current_buffer
    document = buffer.document
    selection_state = buffer.selection_state
    lines = buffer.text.splitlines(keepends=False)

    start, end = (
        document.translate_index_to_position(x)[0] for x in document.selection_range()
    )
    cursor_in_first_line = document.cursor_position_row == start
    # Only remove comments if all lines in the selection have a comment
    uncommenting = all(
        line.lstrip().startswith(comment.rstrip()) for line in lines[start : end + 1]
    )
    if uncommenting:
        for i in range(start, end + 1):
            if not lines:
                lines = [""]
            # Replace the first instance of the comment in each line
            lines[i] = lines[i].replace(comment, "", 1)
            if len(lines[i]) < len(comment):
                # The line might be blank and have trailing whitespace removed
                lines[i] = lines[i].replace(comment.rstrip(), "", 1)
        # Find cursor and selection column positions
        cur_col = document.translate_index_to_position(buffer.cursor_position)[1]
        if selection_state is not None:
            sel_col = document.translate_index_to_position(
                selection_state.original_cursor_position
            )[1]
        # Move cursor & adjust selection
        if cursor_in_first_line:
            bcp = buffer.cursor_position - min(len(comment), cur_col)
            if selection_state is not None:
                selection_state.original_cursor_position -= len(comment) * (
                    end - start
                ) + min(len(comment), sel_col)
        else:
            if selection_state is not None:
                selection_state.original_cursor_position -= min(len(comment), sel_col)
            bcp = buffer.cursor_position - (
                len(comment) * (end - start) + min(len(comment), cur_col)
            )
    else:
        # Find the minimum leading whitespace in the selected lines
        whitespace = min(
            len(line) - len(line.lstrip()) for line in lines[start : end + 1]
        )
        for i in range(start, end + 1):
            # Add a comment after the minimum leading whitespace to each line
            line = lines[i]
            lines[i] = line[:whitespace] + comment + line[whitespace:]
        # Move cursor & adjust selection
        if cursor_in_first_line:
            bcp = buffer.cursor_position + len(comment)
            if selection_state is not None:
                selection_state.original_cursor_position += len(comment) * (
                    end - start + 1
                )
        else:
            if selection_state is not None:
                selection_state.original_cursor_position += len(comment)
            bcp = buffer.cursor_position + len(comment) * (end - start + 1)

    # Set the buffer text, curor position and selection state
    buffer.document = Document("\n".join(lines), bcp)
    buffer.selection_state = selection_state


def wrap_selection_cmd(left: str, right: str) -> None:
    """Add strings to either end of the current selection."""
    buffer = get_app().current_buffer
    selection_state = buffer.selection_state
    for start, end in buffer.document.selection_ranges():
        buffer.transform_region(start, end, lambda s: f"{left}{s}{right}")
    # keep the selection of the inner expression
    buffer.cursor_position += len(left)
    if selection_state is not None:
        selection_state.original_cursor_position += len(left)
    buffer.selection_state = selection_state


WRAP_PAIRS: dict[str, list[str]] = {
    "code": [
        '""',
        "''",
        "()",
        "{}",
        "[]",
    ],
    "markdown": [
        "``",
        "**",
        "__",
        "<>",
    ],
}

for pair in WRAP_PAIRS["code"]:
    left, right = list(pair)
    add_cmd(
        name=f"wrap-selection-{pair}",
        title=f"Wrap selection in {pair}",
        description=f"Wraps the current selection with: {pair}",
        filter=buffer_has_focus & has_selection & buffer_is_code,
    )(partial(wrap_selection_cmd, left, right))


for pair in WRAP_PAIRS["markdown"]:
    left, right = list(pair)
    add_cmd(
        name=f"wrap-selection-{pair}",
        # keys=sorted(set(pair)),
        title=f"Wrap selection in {pair}",
        description=f"Wraps the current selection with: {pair}",
        filter=buffer_has_focus & has_selection & buffer_is_markdown,
    )(partial(wrap_selection_cmd, left, right))


@add_cmd(
    filter=buffer_has_focus & ~has_selection,
)
def duplicate_line() -> None:
    """Duplicate the current line."""
    buffer = get_app().current_buffer
    line = buffer.document.current_line
    eol = buffer.document.get_end_of_line_position()
    buffer.cursor_position += eol
    buffer.newline(copy_margin=False)
    buffer.insert_text(line)
    buffer.cursor_position -= eol


@add_cmd(
    filter=buffer_has_focus & has_selection,
)
def duplicate_selection() -> None:
    """Duplicate the current selection."""
    buffer = get_app().current_buffer
    selection_state = buffer.selection_state
    from_, to = buffer.document.selection_range()
    text = buffer.document.text[from_:to]
    buffer.insert_text(text)
    buffer.selection_state = selection_state


@add_cmd(
    title="Paste",
    filter=buffer_has_focus,
)
def paste_clipboard() -> None:
    """Pate the clipboard contents, replacing any current selection."""
    app = get_app()
    buff = app.current_buffer
    if buff.selection_state:
        buff.cut_selection()
    buff.paste_clipboard_data(app.clipboard.get_data())


@add_cmd(
    title="Copy",
    filter=has_selection,
)
def copy_selection() -> None:
    """Add the current selection to the clipboard."""
    app = get_app()
    buffer = app.current_buffer
    selection_state = buffer.selection_state
    data = buffer.copy_selection()
    buffer.selection_state = selection_state
    app.clipboard.set_data(data)


@add_cmd(
    title="Cut",
    filter=has_selection,
)
def cut_selection() -> None:
    """Remove the current selection and adds it to the clipboard."""
    data = get_app().current_buffer.cut_selection()
    get_app().clipboard.set_data(data)


@add_cmd(
    filter=buffer_has_focus,
)
def cut_line() -> None:
    """Remove the current line adds it to the clipboard."""
    app = get_app()
    buffer = app.current_buffer
    clipboard = app.clipboard
    clipboard_text = clipboard.get_data().text
    line = buffer.document.current_line
    clipboard.set_text(f"{clipboard_text}\n{line}")
    lines = buffer.document.lines[:]
    del lines[buffer.document.cursor_position_row]
    text = "\n".join(lines)
    buffer.document = Document(
        text,
        min(
            buffer.cursor_position + buffer.document.get_start_of_line_position(),
            len(text),
        ),
    )


def move_line(n: int) -> None:
    """Move the current or selected lines up or down by one or more lines."""
    buffer = get_app().current_buffer
    selection_state = buffer.selection_state
    lines = buffer.text.splitlines(keepends=False)
    start, end = (
        buffer.document.translate_index_to_position(x)[0]
        for x in buffer.document.selection_range()
    )
    end += 1
    # Check that we are not moving lines off the edge of the document
    if start + n < 0 or end > len(lines):
        return
    # Rearrange the lines
    sel_lines = lines[start:end]
    lines_new = lines[:start] + lines[end:]
    lines_new = lines_new[: start + n] + sel_lines + lines_new[start + n :]
    # Calculate the new cursor position
    row = buffer.document.cursor_position_row
    col = buffer.document.cursor_position_col
    text = "\n".join(lines_new)
    cursor_position_new = Document(text).translate_row_col_to_index(row + n, col)
    cursor_position_diff = buffer.cursor_position - cursor_position_new
    # Update the selection if we have one
    if selection_state:
        selection_state.original_cursor_position -= cursor_position_diff
    # Update the buffer contents
    buffer.document = Document(text, cursor_position_new)
    buffer.selection_state = selection_state


@add_cmd(
    filter=buffer_has_focus,
)
def move_lines_up() -> None:
    """Move the current or selected lines up by one line."""
    move_line(-1)


@add_cmd(
    filter=buffer_has_focus,
)
def move_lines_down() -> None:
    """Move the current or selected lines down by one line."""
    move_line(1)


add_cmd(
    filter=insert_mode & is_returnable & ~is_multiline,
    description="Accept an input.",
)(accept_line)


def dent_buffer(event: KeyPressEvent, indenting: bool = True) -> None:
    """Indent or unindent the current or selected lines in a buffer."""
    buffer = get_app().current_buffer
    document = buffer.document
    selection_state = buffer.selection_state
    cursor_position = buffer.cursor_position
    lines = buffer.document.lines

    # Apply indentation end the selected range
    start, end = (
        document.translate_index_to_position(x)[0] for x in document.selection_range()
    )
    dent_func = indent if indenting else unindent
    dent_func(buffer, start, end + 1)

    # If there is a selection, adjust the selection range
    if selection_state:
        # Calculate the direction of the change
        sign = indenting * 2 - 1
        # Determine which lines were affected
        lines_affected = {
            i: indenting or lines[i][:1] == " " for i in range(start, end + 1)
        }
        total_lines_affected = sum(lines_affected.values())
        # If the cursor was in the first line, it will only move by the change of that line
        # otherwise it will move by the total change of all lines
        cursor_in_first_line = document.cursor_position_row == start
        diffs = (
            get_app().config.tab_size * total_lines_affected,
            get_app().config.tab_size * lines_affected[start],
        )
        # Do not move the cursor or the start of the selection back onto a previous line
        buffer.cursor_position = (
            cursor_position
            + max(
                diffs[cursor_in_first_line],
                document.get_start_of_line_position(),
            )
            * sign
        )
        og_cursor_col = document.translate_index_to_position(
            selection_state.original_cursor_position
        )[1]
        selection_state.original_cursor_position += (
            max(
                diffs[not cursor_in_first_line],
                -og_cursor_col,
            )
            * sign
        )
    # Maintain the selection state before indentation
    buffer.selection_state = selection_state


@add_cmd(
    filter=buffer_has_focus & is_multiline,
)
def newline(event: KeyPressEvent) -> None:
    """Inert a new line, replacing any selection and indenting if appropriate."""
    # TODO https://git.io/J9GfI
    buffer = get_app().current_buffer
    document = buffer.document
    buffer.cut_selection()
    buffer.newline(copy_margin=not in_paste_mode())
    if buffer_is_code():
        pre = document.current_line_before_cursor.rstrip()
        post = document.current_line_after_cursor.lstrip()
        if post and post.strip()[0] in (")", "]", "}"):
            buffer.insert_text(
                "\n" + buffer.document.leading_whitespace_in_current_line,
                move_cursor=False,
                fire_event=False,
            )
        if pre and pre[-1:] in (":", "(", "[", "{"):
            dent_buffer(event)


@add_cmd(
    filter=(buffer_has_focus & (cursor_in_leading_ws | has_selection)),
)
def indent_lines(event: KeyPressEvent) -> None:
    """Inndent the current or selected lines."""
    dent_buffer(event)


@add_cmd(
    name="unindent-line",
    filter=cursor_in_leading_ws & ~has_selection & ~cursor_at_start_of_line,
)
@add_cmd(
    filter=buffer_has_focus
    & (cursor_in_leading_ws | has_selection)
    & ~cursor_at_start_of_line,
)
def unindent_lines(event: KeyPressEvent) -> None:
    """Unindent the current or selected lines."""
    dent_buffer(event, indenting=False)


@add_cmd(
    filter=buffer_has_focus,
)
def toggle_case() -> None:
    """Toggle the case of the current word or selection."""
    buffer = get_app().current_buffer
    selection_state = buffer.selection_state
    if selection_state is None:
        start, end = buffer.document.find_boundaries_of_current_word()
        selection_state = SelectionState(buffer.cursor_position + start)
        selection_state.enter_shift_mode()
        buffer.cursor_position += end
    if selection_state is not None:
        cp = buffer.cursor_position
        text = buffer.cut_selection().text
        if text.islower():
            text = text.title()
        elif text.istitle():
            text = text.upper()
        else:
            text = text.lower()
        buffer.insert_text(text)
        buffer.cursor_position = cp
        buffer.selection_state = selection_state


@add_cmd(
    filter=buffer_has_focus,
)
def undo() -> None:
    """Undo the last edit."""
    get_app().current_buffer.undo()


@add_cmd(
    filter=buffer_has_focus,
)
def redo() -> None:
    """Redo the last edit."""
    get_app().current_buffer.redo()


# Selection


@add_cmd(
    filter=buffer_has_focus,
)
def select_all() -> None:
    """Select all text."""
    buffer = get_app().current_buffer
    buffer.selection_state = SelectionState(0)
    buffer.cursor_position = len(buffer.text)
    buffer.selection_state.enter_shift_mode()


def unshift_move(event: KeyPressEvent) -> None:
    """Handle keys in shift selection mode.

    When called with a shift + movement key press event, moves the cursor as if shift
    is not pressed.

    Args:
        event: The key press event to process

    """
    key = event.key_sequence[0].key

    if key == Keys.ShiftUp:
        event.current_buffer.auto_up(count=event.arg)
        return
    if key == Keys.ShiftDown:
        event.current_buffer.auto_down(count=event.arg)
        return

    # the other keys are handled through their readline command
    key_to_command: dict[tuple[Keys | str, ...], str] = {
        (Keys.ShiftLeft,): "move-cursor-left",
        (Keys.ShiftRight,): "move-cursor-right",
        (Keys.ShiftHome,): "go-to-start-of-line",
        (Keys.ShiftEnd,): "go-to-end-of-line",
        (
            Keys.Escape,
            Keys.ShiftLeft,
        ): "backward-word",
        (
            Keys.Escape,
            Keys.ShiftRight,
        ): "forward-word",
        (Keys.ControlShiftLeft,): "backward-word",
        (Keys.ControlShiftRight,): "forward-word",
        (Keys.ControlShiftHome,): "beginning-of-buffer",
        (Keys.ControlShiftEnd,): "end-of-buffer",
    }

    key_sequence = tuple(key_ress.key for key_ress in event.key_sequence)
    try:
        command = get_cmd(key_to_command[key_sequence])
    except KeyError:
        pass
    else:
        command.key_handler(event)


@add_cmd(filter=~has_selection, hidden=True)
def start_selection(event: KeyPressEvent) -> None:
    """Start a new selection."""
    # Take the current cursor position as the start of this selection.
    buff = event.current_buffer
    if buff.text:
        buff.start_selection(selection_type=SelectionType.CHARACTERS)
        if buff.selection_state is not None:
            buff.selection_state.enter_shift_mode()

        # Then move the cursor
        original_position = buff.cursor_position
        unshift_move(event)
        if buff.cursor_position == original_position:
            # Cursor didn't actually move - so cancel selection
            # to avoid having an empty selection
            buff.exit_selection()


@add_cmd(filter=shift_selection_mode, hidden=True)
def extend_selection(event: KeyPressEvent) -> None:
    """Extend the selection."""
    # Just move the cursor, like shift was not pressed
    unshift_move(event)
    buff = event.current_buffer
    if buff.selection_state is not None:
        if buff.cursor_position == buff.selection_state.original_cursor_position:
            # selection is now empty, so cancel selection
            buff.exit_selection()


@add_cmd(filter=has_selection, hidden=True)
def replace_selection(event: KeyPressEvent) -> None:
    """Replace selection by what is typed."""
    event.current_buffer.cut_selection()
    get_by_name("self-insert").call(event)


@add_cmd(filter=has_selection, hidden=True)
def delete_selection() -> None:
    """Delete the contents of the current selection."""
    get_app().current_buffer.cut_selection()


@add_cmd(filter=shift_selection_mode, hidden=True)
def cancel_selection(event: KeyPressEvent) -> None:
    """Cancel the selection."""
    event.current_buffer.exit_selection()
    # we then process the cursor movement
    key_press = event.key_sequence[0]
    event.key_processor.feed(key_press, first=True)


@add_cmd(
    filter=buffer_has_focus,
)
def go_to_matching_bracket(event: KeyPressEvent) -> None:
    """Go to matching bracket if the cursor is on a paired bracket."""
    buff = event.current_buffer
    buff.cursor_position += buff.document.find_matching_bracket_position()


@add_cmd(
    filter=has_suggestion,
)
def accept_suggestion(event: KeyPressEvent) -> None:
    """Accept suggestion."""
    b = get_app().current_buffer
    suggestion = b.suggestion
    if suggestion:
        b.insert_text(suggestion.text)


@add_cmd(
    filter=has_suggestion,
)
def fill_sugestion(event: KeyPressEvent) -> None:
    """Fill partial suggestion."""
    b = get_app().current_buffer
    suggestion = b.suggestion
    if suggestion:
        t = re.split(r"(\S+\s+)", suggestion.text)
        b.insert_text(next(x for x in t if x))



"""Initiate logging for euporie.core."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition, has_focus
from prompt_toolkit.formatted_text.base import FormattedText
from prompt_toolkit.layout.containers import HSplit, VSplit
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.widgets import SearchToolbar

from euporie.core.commands import add_cmd
from euporie.core.current import get_app
from euporie.core.log import LOG_QUEUE, QueueHandler
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.path import parse_path
from euporie.core.tabs.base import Tab
from euporie.core.widgets.formatted_text_area import FormattedTextArea

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Callable

    from euporie.core.app import BaseApp


class LogView(Tab):
    """A tab which allows you to view log entries."""

    def __init__(self, app: BaseApp, path: Path | None = None) -> None:
        """Create a new log view tab instance."""
        super().__init__(app, path)
        # Build the container
        self.search_field = SearchToolbar(
            text_if_not_searching=[("class:not-searching", "Press '/' to search.")]
        )
        self.text_area = FormattedTextArea(
            formatted_text=[],
            read_only=True,
            scrollbar=False,
            line_numbers=Condition(lambda: self.app.config.line_numbers),
            search_field=self.search_field,
            focus_on_click=True,
            wrap_lines=False,
            dont_extend_width=False,
        )
        self.container = HSplit(
            [
                VSplit(
                    [
                        self.text_area,
                        MarginContainer(
                            ScrollbarMargin(), target=self.text_area.window
                        ),
                    ]
                ),
                self.search_field,
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
        )
        self.hook_id = QueueHandler.hook(self.add_record)
        # Add text to the textarea
        for record in LOG_QUEUE:
            self.add_record(record)
        # Hook the queue handler
        self.text_area.window.cursorline = has_focus(self)

    def add_record(self, message: FormattedText) -> None:
        """Add a single new record to the textarea.

        Args:
            message: The formatted log record to add

        """
        cp = self.text_area.buffer.cursor_position
        self.text_area.formatted_text += message
        self.text_area.buffer.cursor_position = cp

    @property
    def title(self) -> str:
        """Return the title of this tab."""
        suffix = (
            f" ({parse_path(self.app.config.log_file).name})"
            if self.app.config.log_file
            else ""
        )
        return f"Logs{suffix}"

    def close(self, cb: Callable | None = None) -> None:
        """Remove log queue handler hook on close."""
        QueueHandler.unhook(self.hook_id)
        super().close(cb=cb)

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _view_logs() -> None:
        """Open the logs in a new tab."""
        app = get_app()
        for tab in app.tabs:
            if isinstance(tab, LogView):
                break
        else:
            tab = LogView(app)
            app.tabs.append(tab)
        tab.focus()



"""Key bindings to deal with pixel mouse positioning."""

import logging
from typing import TYPE_CHECKING, FrozenSet, NamedTuple

from prompt_toolkit.data_structures import Point
from prompt_toolkit.key_binding.bindings.mouse import (
    MOUSE_MOVE,
    UNKNOWN_BUTTON,
    UNKNOWN_MODIFIER,
    KeyBindings,
    typical_mouse_events,
    urxvt_mouse_events,
    xterm_sgr_mouse_events,
)
from prompt_toolkit.key_binding.bindings.mouse import (
    load_mouse_bindings as load_ptk_mouse_bindings,
)
from prompt_toolkit.keys import Keys
from prompt_toolkit.mouse_events import MouseButton, MouseEventType, MouseModifier
from prompt_toolkit.mouse_events import MouseEvent as PtkMouseEvent

from euporie.core.app import BaseApp

if TYPE_CHECKING:
    from typing import Optional

    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

log = logging.getLogger(__name__)


class RelativePosition(NamedTuple):
    """Store the relative position or the mouse within a terminal cell."""

    x: float
    y: float


class MouseEvent(PtkMouseEvent):
    """Mouse event, which also store relative position of the mouse event in a cell."""

    def __init__(
        self,
        position: "Point",
        event_type: "MouseEventType",
        button: "MouseButton",
        modifiers: "FrozenSet[MouseModifier]",
        cell_position: "Optional[RelativePosition]",
    ) -> "None":
        """Create new event instance."""
        super().__init__(
            position=position,
            event_type=event_type,
            button=button,
            modifiers=modifiers,
        )
        self.cell_position = cell_position or RelativePosition(0.5, 0.5)


def load_mouse_bindings() -> "KeyBindings":
    """Additional key-bindings to deal with SGR-pixel mouse positioning."""
    key_bindings = load_ptk_mouse_bindings()

    @key_bindings.add(Keys.Vt100MouseEvent)
    def _(event: "KeyPressEvent") -> "NotImplementedOrNone":
        """Handle incoming mouse event, include SGR-pixel mode."""
        # Ensure mypy knows this would only run in a euporie app
        assert isinstance(event.app, BaseApp)

        rx = ry = 0.5

        # Parse incoming packet.
        if event.data[2] == "M":
            # Typical.
            mouse_event, x, y = map(ord, event.data[3:])

            # TODO: Is it possible to add modifiers here?
            mouse_button, mouse_event_type, mouse_modifiers = typical_mouse_events[
                mouse_event
            ]

            # Handle situations where `PosixStdinReader` used surrogateescapes.
            if x >= 0xDC00:
                x -= 0xDC00
            if y >= 0xDC00:
                y -= 0xDC00

            x -= 32
            y -= 32
        else:
            # Urxvt and Xterm SGR.
            # When the '<' is not present, we are not using the Xterm SGR mode,
            # but Urxvt instead.
            data = event.data[2:]
            if data[:1] == "<":
                sgr = True
                data = data[1:]
            else:
                sgr = False

            # Extract coordinates.
            mouse_event, x, y = map(int, data[:-1].split(";"))
            m = data[-1]

            # Parse event type.
            if sgr:
                if event.app.term_info.sgr_pixel_status.value:
                    # Calculate cell position
                    cell_px, cell_py = event.app.term_info.cell_size_px
                    px, py = x, y
                    fx, fy = px / cell_px + 1, py / cell_py + 1
                    x, y = int(fx), int(fy)
                    rx, ry = fx - x, fy - y

                try:
                    (
                        mouse_button,
                        mouse_event_type,
                        mouse_modifiers,
                    ) = xterm_sgr_mouse_events[mouse_event, m]
                except KeyError:
                    return NotImplemented

            else:
                # Some other terminals, like urxvt, Hyper terminal, ...
                (
                    mouse_button,
                    mouse_event_type,
                    mouse_modifiers,
                ) = urxvt_mouse_events.get(
                    mouse_event, (UNKNOWN_BUTTON, MOUSE_MOVE, UNKNOWN_MODIFIER)
                )

        x -= 1
        y -= 1

        # Only handle mouse events when we know the window height.
        if event.app.renderer.height_is_known and mouse_event_type is not None:
            # Take region above the layout into account. The reported
            # coordinates are absolute to the visible part of the terminal.
            from prompt_toolkit.renderer import HeightIsUnknownError

            try:
                y -= event.app.renderer.rows_above_layout
            except HeightIsUnknownError:
                return NotImplemented

            # Save global mouse position
            event.app.mouse_position = Point(x=x, y=y)

            # Apply limits to mouse position if enabled
            if (mouse_limits := event.app.mouse_limits) is not None:
                x = max(
                    mouse_limits.xpos,
                    min(x, mouse_limits.xpos + (mouse_limits.width - 1)),
                )
                y = max(
                    mouse_limits.ypos,
                    min(y, mouse_limits.ypos + (mouse_limits.height - 1)),
                )

            # Call the mouse handler from the renderer.
            # Note: This can return `NotImplemented` if no mouse handler was
            #       found for this position, or if no repainting needs to
            #       happen. this way, we avoid excessive repaints during mouse
            #       movements.
            handler = event.app.renderer.mouse_handlers.mouse_handlers[y][x]

            return handler(
                MouseEvent(
                    position=Point(x=x, y=y),
                    event_type=mouse_event_type,
                    button=mouse_button,
                    modifiers=mouse_modifiers,
                    cell_position=RelativePosition(rx, ry),
                )
            )

        return NotImplemented

    return key_bindings
